try:
    from . import generic as g
except BaseException:
    import generic as g

from trimesh.scene.transforms import EnforcedForest


def random_chr():
    return chr(ord("a") + round(g.random() * 25))


def test_forest():
    graph = EnforcedForest()
    for _i in range(5000):
        graph.add_edge(random_chr(), random_chr())


def test_cache():
    for _i in range(10):
        scene = g.trimesh.Scene()
        scene.add_geometry(g.trimesh.creation.box())

        mod = [scene.graph.__hash__()]
        scene.set_camera()
        mod.append(scene.graph.__hash__())
        assert mod[-1] != mod[-2]

        assert not g.np.allclose(scene.camera_transform, g.np.eye(4))
        scene.camera_transform = g.np.eye(4)
        mod.append(scene.graph.__hash__())
        assert mod[-1] != mod[-2]

        assert g.np.allclose(scene.camera_transform, g.np.eye(4))
        assert mod[-1] != mod[-2]


def test_successors():
    s = g.get_mesh("CesiumMilkTruck.glb")
    assert len(s.graph.nodes_geometry) == 5

    # world should be root frame
    assert s.graph.transforms.successors(s.graph.base_frame) == set(s.graph.nodes)

    for n in s.graph.nodes:
        # successors should always return subset of nodes
        succ = s.graph.transforms.successors(n)
        assert succ.issubset(s.graph.nodes)
        # we self-include node in successors
        assert n in succ

    # test getting a subscene from successors
    ss = s.subscene("3")
    assert len(ss.geometry) == 1
    assert len(ss.graph.nodes_geometry) == 1

    assert isinstance(s.graph.to_networkx(), g.nx.DiGraph)


def test_nodes():
    # get a scene graph
    graph = g.get_mesh("cycloidal.3DXML").graph
    # get any non-root node
    node = next(iter(set(graph.nodes).difference([graph.base_frame])))
    # remove that node
    graph.transforms.remove_node(node)
    # should have dumped the cache and removed the node
    assert node not in graph.nodes


def test_remove_geometries():
    # remove geometries from a scene graph
    scene = g.get_mesh("cycloidal.3DXML")

    # only keep geometry instances of these
    keep = {"disc_cam_A", "disc_cam_B", "vxb-6800-2rs"}

    assert len(scene.duplicate_nodes) == 12

    # should remove instance references except `keep`
    scene.graph.remove_geometries(set(scene.geometry.keys()).difference(keep))

    # there should now be three groups of duplicate nodes
    assert len(scene.duplicate_nodes) == len(keep)


def test_kwargs():
    # test the function that converts various
    # arguments into a homogeneous transformation
    f = g.trimesh.scene.transforms.kwargs_to_matrix
    # no arguments should be an identity matrix
    assert g.np.allclose(f(), g.np.eye(4))

    # a passed matrix should return immediately
    fix = g.random((4, 4))
    assert g.np.allclose(f(matrix=fix), fix)

    quat = g.trimesh.unitize([1, 2, 3, 1])
    trans = [1.0, 2.0, 3.0]
    rot = g.trimesh.transformations.quaternion_matrix(quat)
    # should be the same as passed to transformations
    assert g.np.allclose(rot, f(quaternion=quat))

    # try passing both quaternion and translation
    combine = f(quaternion=quat, translation=trans)
    # should be the same as passed and computed
    assert g.np.allclose(combine[:3, :3], rot[:3, :3])
    assert g.np.allclose(combine[:3, 3], trans)


def test_remove_node():
    s = g.get_mesh("CesiumMilkTruck.glb")

    assert len(s.graph.nodes_geometry) == 5
    assert len(s.graph.nodes) == 9
    assert len(s.graph.transforms.node_data) == 9
    assert len(s.graph.transforms.edge_data) == 8
    assert len(s.graph.transforms.parents) == 8

    assert s.graph.transforms.remove_node("1")

    assert len(s.graph.nodes_geometry) == 5
    assert len(s.graph.nodes) == 8
    assert len(s.graph.transforms.node_data) == 8
    assert len(s.graph.transforms.edge_data) == 6
    assert len(s.graph.transforms.parents) == 6


def test_subscene():
    s = g.get_mesh("CesiumMilkTruck.glb")

    assert len(s.graph.nodes) == 9
    assert len(s.graph.transforms.node_data) == 9
    assert len(s.graph.transforms.edge_data) == 8

    ss = s.subscene("3")

    assert ss.graph.base_frame == "3"
    assert set(ss.graph.nodes) == {"3", "4"}
    assert len(ss.graph.transforms.node_data) == 2
    assert len(ss.graph.transforms.edge_data) == 1
    assert list(ss.graph.transforms.edge_data.keys()) == [("3", "4")]


def test_scene_transform():
    # get a scene graph
    scene = g.get_mesh("cycloidal.3DXML")

    # copy the original bounds of the scene's convex hull
    b = scene.convex_hull.bounds.tolist()
    # dump it into a single mesh
    m = scene.to_mesh()

    # mesh bounds should match exactly
    assert g.np.allclose(m.bounds, b)
    assert g.np.allclose(scene.convex_hull.bounds, b)

    # get a random rotation matrix
    T = g.trimesh.transformations.random_rotation_matrix()

    # apply it to both the mesh and the scene
    m.apply_transform(T)
    scene.apply_transform(T)

    # the mesh and scene should have the same bounds
    assert g.np.allclose(m.convex_hull.bounds, scene.convex_hull.bounds)
    # should have moved from original position
    assert not g.np.allclose(m.convex_hull.bounds, b)


def test_simplify():
    if not g.trimesh.util.has_module("fast_simplification"):
        return

    # get a scene graph
    scene: g.trimesh.Scene = g.get_mesh("cycloidal.3DXML")

    original = scene.to_mesh()

    scene.simplify_quadric_decimation(percent=0.0, aggression=0)
    assert len(scene.to_mesh().vertices) < len(original.vertices)


def test_reverse():
    tf = g.trimesh.transformations

    s = g.trimesh.scene.Scene()
    s.add_geometry(
        g.trimesh.creation.box(),
        parent_node_name="world",
        node_name="foo",
        transform=tf.translation_matrix([0, 0, 1]),
    )

    s.add_geometry(
        g.trimesh.creation.box(),
        parent_node_name="foo",
        node_name="foo2",
        transform=tf.translation_matrix([0, 0, 1]),
    )

    assert len(s.graph.transforms.edge_data) == 2
    a = s.graph.get(frame_from="world", frame_to="foo2")

    assert len(s.graph.transforms.edge_data) == 2

    # try going backward
    i = s.graph.get(frame_from="foo2", frame_to="world")
    # matrix should be inverted if you're going the other way
    assert g.np.allclose(a[0], g.np.linalg.inv(i[0]))

    # try getting foo2 with shorthand
    b = s.graph.get(frame_to="foo2")
    c = s.graph["foo2"]
    # matrix should be inverted if you're going the other way
    assert g.np.allclose(a[0], c[0])
    assert g.np.allclose(b[0], c[0])

    # get should not have edited edge data
    assert len(s.graph.transforms.edge_data) == 2


def test_shortest_path():
    # compare the EnforcedForest shortest path algo
    # to the more general networkx.shortest_path algo
    if g.sys.version_info < (3, 7):
        # old networkx is a lot different
        return

    tf = g.trimesh.transformations
    # start with a known good random tree
    edges = [tuple(row) for row in g.data["random_tree"]]
    tree = g.nx.from_edgelist(edges, create_using=g.nx.DiGraph)

    r_choices = g.random((len(edges), 2))
    r_matrices = g.random_transforms(len(edges))
    edgelist = {}
    for e, r_choice, r_mat in zip(edges, r_choices, r_matrices):
        data = {}
        if r_choice[0] > 0.5:
            # if a matrix is omitted but an edge exists it is
            # the same as passing an identity matrix
            data["matrix"] = r_mat
        if r_choice[1] > 0.4:
            # a geometry is not required for a node
            data["geometry"] = str(int(r_choice[1] * 1e8))
        edgelist[e] = data

    # now apply the random data to an EnforcedForest
    forest = g.trimesh.scene.transforms.EnforcedForest()
    for k, v in edgelist.items():
        forest.add_edge(*k, **v)

    # generate a lot of random queries
    queries = g.np.random.choice(list(forest.nodes), 10000).reshape((-1, 2))
    # filter out any self-queries as networkx doesn't handle them
    queries = queries[g.np.ptp(queries, axis=1) > 0]

    # now run our shortest path algorithm in a profiler
    with g.Profiler() as P:
        ours = [forest.shortest_path(*q) for q in queries]
    # print this way to avoid a python2 syntax error
    g.log.debug(P.output_text())

    # check truth from networkx with an undirected graph
    undir = tree.to_undirected()
    with g.Profiler() as P:
        truth = [g.nx.shortest_path(undir, *q) for q in queries]
    g.log.debug(P.output_text())

    # now compare our shortest path with networkx
    for a, b, q in zip(truth, ours, queries):
        if tuple(a) != tuple(b):
            # raise the query that killed us
            raise ValueError(q)

    # now try creating this as a full scenegraph
    sg = g.trimesh.scene.transforms.SceneGraph()
    [
        sg.update(frame_from=k[0], frame_to=k[1], **kwargs)
        for k, kwargs in edgelist.items()
    ]

    with g.Profiler() as P:
        matgeom = [sg.get(frame_from=q[0], frame_to=q[1]) for q in queries]
    g.log.debug(P.output_text())

    # all of the matrices should be rigid transforms
    assert all(tf.is_rigid(mat) for mat, _ in matgeom)


def test_scaling_order():
    s = g.trimesh.creation.box().scene()
    scaling = 1.0 / 3.0
    c = s.scaled(scaling)
    factor = c.geometry["geometry_0"].vertices / s.geometry["geometry_0"].vertices
    assert g.np.allclose(factor, scaling)
    # should be returning itself
    r = s.apply_translation([10.5, 10.5, 10.5])
    assert g.np.allclose(r.bounds, [[10, 10, 10], [11, 11, 11]])
    assert g.np.allclose(s.bounds, [[10, 10, 10], [11, 11, 11]])


def test_translation_cache():
    # scene with non-geometry nodes
    c = g.get_mesh("cycloidal.3DXML")
    s = c.scaled(1.0 / c.extents)
    # get the pre-translation bounds
    ori = s.bounds.copy()
    # apply a translation
    s.apply_translation([10, 10, 10])
    assert g.np.allclose(s.bounds, ori + 10)


def test_translation_origin():
    # check to see if we can translate to the origin
    c = g.get_mesh("cycloidal.3DXML")
    c.apply_transform(g.trimesh.transformations.random_rotation_matrix())
    s = c.scaled(1.0 / c.extents)
    # shouldn't be at the origin
    assert not g.np.allclose(s.bounds[0], 0.0)
    # should move to the origin
    s.apply_translation(-s.bounds[0])
    assert g.np.allclose(s.bounds[0], 0)


def test_reconstruct():
    original = g.get_mesh("cycloidal.3DXML")
    assert isinstance(original, g.trimesh.Scene)

    # get the scene as "baked" meshes with no scene graph
    dupe = g.trimesh.Scene(original.dump())
    assert len(dupe.geometry) > len(original.geometry)

    with g.Profiler() as P:
        # reconstruct the instancing using `duplicate_nodes` and `procrustes`
        rec = dupe.reconstruct_instances()
    g.log.info(P.output_text())

    assert len(rec.graph.nodes_geometry) == len(original.graph.nodes_geometry)
    assert len(rec.geometry) == len(original.geometry)
    assert g.np.allclose(rec.extents, original.extents, rtol=1e-8)
    assert g.np.allclose(rec.center_mass, original.center_mass, rtol=1e-8)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()

    test_reconstruct()



try:
    from . import generic as g
except BaseException:
    import generic as g


class VoxelGridTest(g.unittest.TestCase):
    def test_voxel(self):
        """
        Test that voxels work at all
        """
        for m in [
            g.get_mesh("featuretype.STL", force="mesh"),
            g.trimesh.primitives.Box(),
            g.trimesh.primitives.Sphere(),
        ]:
            for pitch in [0.1, 0.1 - g.tol.merge]:
                surface = m.voxelized(pitch=pitch)

                scene = g.trimesh.Scene(surface)
                assert len(scene.geometry) == 1
                assert g.np.allclose(scene.bounds, surface.bounds)

                # make sure the voxelized pitch is similar to passed
                assert g.np.allclose(surface.pitch, pitch)

                for fill_method in ("base", "orthographic"):
                    solid = surface.copy().fill(method=fill_method)

                    assert len(surface.encoding.dense.shape) == 3
                    assert surface.shape == surface.encoding.dense.shape
                    assert surface.volume > 0.0

                    assert isinstance(surface.filled_count, int)
                    assert surface.filled_count > 0

                    box_surface = surface.as_boxes()
                    box_solid = solid.as_boxes()

                    assert isinstance(box_surface, g.trimesh.Trimesh)
                    assert abs(box_solid.volume - solid.volume) < g.tol.merge

                    assert g.trimesh.util.is_shape(surface.sparse_indices, (-1, 3))
                    assert len(solid.sparse_indices) >= len(surface.sparse_indices)
                    assert solid.sparse_indices.shape == solid.points.shape
                    outside = m.bounds[1] + m.scale
                    for vox in surface, solid:
                        assert vox.sparse_indices.shape == vox.points.shape
                        assert g.np.all(vox.is_filled(vox.points))
                        assert not vox.is_filled(outside)

                    try:
                        cubes = surface.marching_cubes
                        assert cubes.area > 0.0
                    except ImportError:
                        g.log.info("no skimage, skipping marching cubes test")

                    g.log.info(
                        "Mesh volume was %f, voxelized volume was %f",
                        m.volume,
                        surface.volume,
                    )

    def test_marching(self):
        """
        Test marching cubes on a matrix
        """
        try:
            from skimage import measure  # NOQA
        except ImportError:
            g.log.warning("no skimage, skipping marching cubes test")
            return

        march = g.trimesh.voxel.ops.matrix_to_marching_cubes

        # make sure offset is correct
        matrix = g.np.ones((3, 3, 3), dtype=bool)
        mesh = march(matrix=matrix)
        assert mesh.is_watertight

        mesh = march(matrix=matrix).apply_scale(3.0)
        assert mesh.is_watertight

        # try an array full of a small number
        matrix = g.np.full((3, 3, 3), 0.01, dtype=g.np.float64)
        # set some to zero
        matrix[:2, :2, :2] = 0.0

        a = march(matrix)
        assert a.is_watertight

        # but above the threshold it should be not empty
        b = march(matrix, threshold=-0.001)
        assert b.is_watertight
        assert b.volume > a.volume

    def test_marching_points(self):
        """
        Try marching cubes on points
        """
        try:
            from skimage import measure  # NOQA
        except ImportError:
            g.log.warning("no skimage, skipping marching cubes test")
            return

        # get some points on the surface of an icosahedron
        points = g.trimesh.creation.icosahedron().sample(1000)
        # make the pitch proportional to scale
        pitch = g.np.ptp(points, axis=0).min() / 10
        # run marching cubes
        mesh = g.trimesh.voxel.ops.points_to_marching_cubes(points=points, pitch=pitch)

        # mesh should have faces
        assert len(mesh.faces) > 0
        # mesh should be roughly centered
        assert (mesh.bounds[0] < -0.5).all()
        assert (mesh.bounds[1] > 0.5).all()

    def test_local(self):
        """
        Try calling local voxel functions
        """
        from trimesh.voxel import creation

        mesh = g.trimesh.creation.box()

        # it should have some stuff
        voxel = creation.local_voxelize(
            mesh=mesh, point=[0.5, 0.5, 0.5], pitch=0.1, radius=5, fill=True
        )

        assert len(voxel.shape) == 3

        # try it when it definitely doesn't hit anything
        empty = creation.local_voxelize(
            mesh=mesh, point=[10, 10, 10], pitch=0.1, radius=5, fill=True
        )
        # shouldn't have hit anything
        assert empty is None

        # try it when it is in the center of a volume
        creation.local_voxelize(
            mesh=mesh, point=[0, 0, 0], pitch=0.1, radius=2, fill=True
        )

    def test_points_to_from_indices(self):
        # indices = (points - origin) / pitch
        points = [[0, 0, 0], [0.04, 0.55, 0.39]]
        origin = [0, 0, 0]
        pitch = 0.1
        indices = [[0, 0, 0], [0, 6, 4]]

        # points -> indices
        indices2 = g.trimesh.voxel.ops.points_to_indices(
            points=points, origin=origin, pitch=pitch
        )

        g.np.testing.assert_allclose(indices, indices2, atol=0, rtol=0)

        # indices -> points
        points2 = g.trimesh.voxel.ops.indices_to_points(
            indices=indices, origin=origin, pitch=pitch
        )
        g.np.testing.assert_allclose(
            g.np.array(indices) * pitch + origin, points2, atol=0, rtol=0
        )
        g.np.testing.assert_allclose(points, points2, atol=pitch / 2 * 1.01, rtol=0)

        # indices -> points -> indices (this must be consistent)
        points2 = g.trimesh.voxel.ops.indices_to_points(
            indices=indices, origin=origin, pitch=pitch
        )
        indices2 = g.trimesh.voxel.ops.points_to_indices(
            points=points2, origin=origin, pitch=pitch
        )
        g.np.testing.assert_allclose(indices, indices2, atol=0, rtol=0)

    def test_as_boxes(self):
        voxel = g.trimesh.voxel

        pitch = 0.1
        origin = [0, 0, 1]

        matrix = g.np.eye(9, dtype=bool).reshape((-1, 3, 3))
        centers = g.trimesh.voxel.ops.matrix_to_points(
            matrix=matrix, pitch=pitch, origin=origin
        )
        v = voxel.VoxelGrid(matrix).apply_scale(pitch).apply_translation(origin)

        boxes1 = v.as_boxes()
        boxes2 = g.trimesh.voxel.ops.multibox(centers).apply_scale(pitch)
        colors = [g.trimesh.visual.DEFAULT_COLOR] * matrix.sum() * 12
        for boxes in [boxes1, boxes2]:
            g.np.testing.assert_allclose(boxes.visual.face_colors, colors, atol=0, rtol=0)

        # check assigning a single color
        color = [255, 0, 0, 255]
        boxes1 = v.as_boxes(colors=color)
        boxes2 = g.trimesh.voxel.ops.multibox(centers=centers, colors=color).apply_scale(
            pitch
        )
        colors = g.np.array([color] * len(centers) * 12)
        for boxes in [boxes1, boxes2]:
            g.np.testing.assert_allclose(boxes.visual.face_colors, colors, atol=0, rtol=0)

        # check matrix colors
        colors = color * g.np.ones(g.np.append(v.shape, 4), dtype=g.np.uint8)
        boxes = v.as_boxes(colors=colors)
        assert g.np.allclose(boxes.visual.face_colors, color, atol=0, rtol=0)

    def test_is_filled(self):
        """More rigorous test of VoxelGrid.is_filled."""
        n = 10
        matrix = g.np.random.uniform(size=(n + 1,) * 3) > 0.5
        not_matrix = g.np.logical_not(matrix)
        pitch = 1.0 / n
        origin = g.np.random.uniform(size=(3,))
        vox = g.trimesh.voxel.VoxelGrid(matrix)
        vox = vox.apply_scale(pitch).apply_translation(origin)
        not_vox = g.trimesh.voxel.VoxelGrid(not_matrix)
        not_vox = not_vox.apply_scale(pitch).apply_translation(origin)
        for a, b in ((vox, not_vox), (not_vox, vox)):
            points = a.points
            # slight jitter - shouldn't change indices
            points += (g.np.random.uniform(size=points.shape) - 1) * 0.4 * pitch
            g.np.random.shuffle(points)

            # all points are filled, and no empty points are filled
            assert g.np.all(a.is_filled(points))
            assert not g.np.any(b.is_filled(points))

            # test different number of dimensions
            points = g.np.stack([points, points[-1::-1]], axis=1)
            assert g.np.all(a.is_filled(points))
            assert not g.np.any(b.is_filled(points))

    def test_vox_sphere(self):
        # should be filled from 0-9
        matrix = g.np.ones((10, 10, 10))
        scale = 0.1
        vox = g.trimesh.voxel.VoxelGrid(matrix).apply_scale(scale)
        # epsilon from zero
        eps = 1e-4
        # should all be contained
        grid = g.trimesh.util.grid_linspace([[eps] * 3, [9 - eps] * 3], 11) * scale
        assert vox.is_filled(grid).all()

        # push it outside the filled area
        grid += 1.0
        assert not vox.is_filled(grid).any()

    def test_roundtrip(self):
        # try exporting and reloading in the "binvox" format
        m = g.trimesh.creation.box()
        v = m.voxelized(pitch=0.1)
        e = v.export(file_type="binvox")
        r = g.trimesh.load(file_obj=g.trimesh.util.wrap_as_stream(e), file_type="binvox")

        assert v.filled_count == r.filled_count
        assert g.np.allclose(r.bounds, v.bounds)

    def _test_equiv(self, v0, v1, query_points=None):
        """
        Test whether or not two `VoxelGrid` representation
        are consistent.

        Tests consistency of:
            shape
            filled_count
            volume
            matrix
            points
            origin
            pitch

        If query_points is provided, also tests
            is_filled
            points_to_indices
            indices_to_points

        Parameters
        ----------
        v0: `VoxelGrid` instance
        v1: `VoxelGrid` instance
        query_points: (optional) points as which `points_to_indices` and
        `is_filled` are tested for consistency.
        """

        def array_as_set(array2d):
            return {tuple(x) for x in array2d}

        # all points are filled
        assert g.np.all(v0.is_filled(v1.points))
        assert g.np.any(v1.is_filled(v0.points))

        # test different number of dimensions
        self.assertEqual(v0.shape, v1.shape)
        self.assertEqual(v0.filled_count, v1.filled_count)
        self.assertEqual(v0.volume, v1.volume)
        g.np.testing.assert_equal(v0.encoding.dense, v1.encoding.dense)
        # points will be in different order, but should contain same coords
        g.np.testing.assert_equal(array_as_set(v0.points), array_as_set(v1.points))
        # g.np.testing.assert_equal(v0.origin, v1.origin)
        # g.np.testing.assert_equal(v0.pitch, v1.pitch)
        if query_points is not None:
            indices0 = v0.points_to_indices(query_points)
            indices1 = v1.points_to_indices(query_points)
            g.np.testing.assert_equal(indices0, indices1)
            g.np.testing.assert_allclose(
                v0.points_to_indices(v0.indices_to_points(indices0)), indices0
            )
            g.np.testing.assert_allclose(
                v1.points_to_indices(v1.indices_to_points(indices1)), indices1
            )
            g.np.testing.assert_equal(
                v0.is_filled(query_points), v1.is_filled(query_points)
            )

    def test_voxel_rle(self):
        from trimesh.voxel import encoding as enc

        np = g.np
        voxel = g.trimesh.voxel
        shape = (4, 4, 4)
        rle_obj = enc.RunLengthEncoding(
            np.array([0, 8, 1, 40, 0, 16], dtype=np.uint8), dtype=bool
        )
        brle_obj = enc.BinaryRunLengthEncoding(np.array([8, 40, 16], dtype=np.uint8))
        v_rle = voxel.VoxelGrid(rle_obj.reshape(shape))
        self.assertEqual(v_rle.filled_count, 40)
        g.np.testing.assert_equal(
            v_rle.encoding.dense, g.np.reshape([0] * 8 + [1] * 40 + [0] * 16, shape)
        )

        v_brle = voxel.VoxelGrid(brle_obj.reshape(shape))
        query_points = g.np.random.uniform(size=(100, 3), high=4)
        self._test_equiv(v_rle, v_brle, query_points)

    def test_hollow(self):
        if not g.has_binvox:
            g.log.warning("no binvox to test!")
            return

        filled = g.trimesh.primitives.Sphere().voxelized(
            pitch=0.1, method="binvox", exact=True
        )
        hollow = filled.copy().hollow()
        self.assertLess(hollow.filled_count, filled.filled_count)
        self.assertGreater(hollow.filled_count, 0)

    def test_fill(self):
        from trimesh.voxel.morphology import fillers

        hollow = g.trimesh.primitives.Sphere().voxelized(pitch=0.1).hollow()

        for key in fillers:
            filled = hollow.copy().fill(key)
            self.assertLess(hollow.filled_count, filled.filled_count)

    def test_strip(self):
        if not g.has_binvox:
            g.log.warning("no binvox to test!")
            return

        octant = g.trimesh.primitives.Sphere().voxelized(
            pitch=0.1, method="binvox", exact=True
        )
        dense = octant.encoding.dense.copy()
        nx, ny, nz = octant.shape
        dense[: nx // 2] = 0
        dense[:, : ny // 2] = 0
        dense[:, :, nz // 2 :] = 0
        octant.encoding = dense
        stripped = octant.copy().strip()
        self.assertEqual(octant.filled_count, stripped.filled_count)
        self.assertEqual(octant.volume, stripped.volume)
        g.np.testing.assert_allclose(octant.points, stripped.points)
        self.assertGreater(octant.encoding.size, stripped.encoding.size)

    def test_binvox_with_dimension(self):
        if not g.has_binvox:
            g.log.warning("no binvox to test!")
            return

        dim = 10
        octant = g.trimesh.primitives.Sphere().voxelized(
            pitch=None, dimension=dim, method="binvox", exact=True
        )
        assert octant.shape == (dim, dim, dim)

    def test_transform_cache(self):
        encoding = [
            [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
            [[0, 1, 1], [0, 1, 0], [1, 1, 0]],
            [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
        ]
        vg = g.trimesh.voxel.VoxelGrid(g.np.asarray(encoding))

        scale = g.np.asarray([12, 23, 24])
        s_matrix = g.np.eye(4)
        s_matrix[:3, :3] *= scale

        # original scale should be identity
        assert g.np.allclose(vg.scale, 1.0)

        # save the hash
        hash_ori = hash(vg._data)
        # modify the voxelgrid
        vg.apply_transform(s_matrix)

        # hash should have changed
        assert hash_ori != hash(vg._data)
        assert g.np.allclose(vg.scale, scale)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class AdjacencyTest(g.unittest.TestCase):
    def test_radius(self):
        for radius in [0.1, 1.0, 3.1459, 29.20]:
            m = g.trimesh.creation.cylinder(radius=radius, height=radius * 10)

            # remove the cylinder cap
            signs = (g.np.sign(m.vertices[:, 2]) < 0)[m.faces]
            not_cap = ~g.np.logical_or(signs.all(axis=1), ~signs.any(axis=1))
            m.update_faces(not_cap)

            # compare the calculated radius
            radii = m.face_adjacency_radius
            radii = radii[g.np.isfinite(radii)]

            assert g.np.allclose(radii, radius, atol=radius / 100)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import numpy as np

from ..constants import res_path as res
from ..constants import tol_path as tol
from ..typed import Integer, List


def discretize_bezier(points, count=None, scale=1.0):
    """
    Parameters
    ----------
    points : (order, dimension) float
      Control points of the bezier curve
      For a 2D cubic bezier, order=3, dimension=2
    count : int, or None
      Number of segments
    scale : float
      Scale of curve
    Returns
    ----------
    discrete: (n, dimension) float
     Points forming a a polyline representation
    """
    # make sure we have a numpy array
    points = np.asanyarray(points, dtype=np.float64)

    if count is None:
        # how much distance does a small percentage of the curve take
        # this is so we can figure out how finely we have to sample t
        norm = np.linalg.norm(np.diff(points, axis=0), axis=1).sum()
        count = np.ceil(norm / (res.seg_frac * scale))
        count = int(
            np.clip(count, res.min_sections * len(points), res.max_sections * len(points))
        )
    count = int(count)

    # parameterize incrementing 0.0 - 1.0
    t = np.linspace(0.0, 1.0, count)
    # decrementing 1.0-0.0
    t_d = 1.0 - t
    n = len(points) - 1
    # binomial coefficients, i, and each point
    iterable = zip(binomial(n), np.arange(len(points)), points)
    # run the actual interpolation
    stacked = [
        ((t**i) * (t_d ** (n - i))).reshape((-1, 1)) * p * c for c, i, p in iterable
    ]
    result = np.sum(stacked, axis=0)

    # a bezier curve always starts and ends on control points
    if tol.strict:
        # test to make sure end points are correct
        test = np.sum((result[[0, -1]] - points[[0, -1]]) ** 2, axis=1)
        assert (test < tol.merge).all()
        assert len(result) >= 2

    # snap the first and last points to the exact control point
    result[[0, -1]] = points[[0, -1]]

    return result


def discretize_bspline(control, knots, count=None, scale=1.0):
    """
    Given a B-Splines control points and knot vector, return
    a sampled version of the curve.

    Parameters
    ----------
    control : (o, d) float
      Control points of the b- spline
    knots : (j,) float
      B-spline knots
    count : int
      Number of line segments to discretize the spline
      If not specified will be calculated as something reasonable

    Returns
    ----------
    discrete : (count, dimension) float
       Points on a polyline version of the B-spline
    """

    # evaluate the b-spline using scipy/fitpack
    from scipy.interpolate import splev

    # (n, d) control points where d is the dimension of vertices
    control = np.asanyarray(control, dtype=np.float64)
    degree = len(knots) - len(control) - 1
    if count is None:
        norm = np.linalg.norm(np.diff(control, axis=0), axis=1).sum()
        count = int(
            np.clip(
                norm / (res.seg_frac * scale),
                res.min_sections * len(control),
                res.max_sections * len(control),
            )
        )

    ipl = np.linspace(knots[0], knots[-1], count)
    discrete = splev(ipl, [knots, control.T, degree])
    discrete = np.column_stack(discrete)

    return discrete


def binomial(n: Integer) -> List:
    """
    Return all binomial coefficients for a given order.

    For n > 5, scipy.special.binom is used, below we hardcode.

    Parameters
    --------------
    n : int
      Order of binomial

    Returns
    ---------------
    binom : (n + 1,) int
      Binomial coefficients of a given order
    """
    if n == 1:
        return [1, 1]
    elif n == 2:
        return [1, 2, 1]
    elif n == 3:
        return [1, 3, 3, 1]
    elif n == 4:
        return [1, 4, 6, 4, 1]
    elif n == 5:
        return [1, 5, 10, 10, 5, 1]
    else:
        from scipy.special import binom

        return binom(n, np.arange(n + 1))



import numpy as np

from ... import graph, grouping, util
from ...constants import tol_path
from ...typed import ArrayLike, Dict, NDArray, Optional
from ..entities import Arc, Line


def dict_to_path(as_dict):
    """
    Turn a pure dict into a dict containing entity objects that
    can be sent directly to a Path constructor.

    Parameters
    ------------
    as_dict : dict
      Has keys: 'vertices', 'entities'

    Returns
    ------------
    kwargs : dict
      Has keys: 'vertices', 'entities'
    """
    # start kwargs with initial value
    result = as_dict.copy()
    # map of constructors
    loaders = {"Arc": Arc, "Line": Line}
    # pre- allocate entity array
    entities = [None] * len(as_dict["entities"])
    # run constructor for dict kwargs
    for entity_index, entity in enumerate(as_dict["entities"]):
        if entity["type"] == "Line":
            entities[entity_index] = loaders[entity["type"]](points=entity["points"])
        else:
            entities[entity_index] = loaders[entity["type"]](
                points=entity["points"], closed=entity["closed"]
            )
    result["entities"] = entities

    return result


def lines_to_path(lines: ArrayLike, index: Optional[NDArray[np.int64]] = None) -> Dict:
    """
    Turn line segments into argument to be used for a Path2D or Path3D.

    Parameters
    ------------
    lines : (n, 2, dimension) or (n, dimension) float
      Line segments or connected polyline curve in 2D or 3D
    index : (n,) int64
      If passed save an index for each line segment.

    Returns
    -----------
    kwargs : Dict
      kwargs for Path constructor
    """
    lines = np.asanyarray(lines, dtype=np.float64)

    if index is not None:
        index = np.asanyarray(index, dtype=np.int64)

    if util.is_shape(lines, (-1, (2, 3))):
        # the case where we have a list of points
        # we are going to assume they are connected
        result = {"entities": np.array([Line(np.arange(len(lines)))]), "vertices": lines}
        return result
    elif util.is_shape(lines, (-1, 2, (2, 3))):
        # case where we have line segments in 2D or 3D
        dimension = lines.shape[-1]
        # convert lines to even number of (n, dimension) points
        lines = lines.reshape((-1, dimension))
        # merge duplicate vertices
        unique, inverse = grouping.unique_rows(lines, digits=tol_path.merge_digits)
        # use scipy edges_to_path to skip creating
        # a bajillion individual line entities which
        # will be super slow vs. fewer polyline entities
        return edges_to_path(edges=inverse.reshape((-1, 2)), vertices=lines[unique])
    else:
        raise ValueError("Lines must be (n,(2|3)) or (n,2,(2|3))")
    return result


def polygon_to_path(polygon):
    """
    Load shapely Polygon objects into a trimesh.path.Path2D object

    Parameters
    -------------
    polygon : shapely.geometry.Polygon
      Input geometry

    Returns
    -----------
    kwargs : dict
      Keyword arguments for Path2D constructor
    """
    # start with a single polyline for the exterior
    entities = []
    # start vertices
    vertices = []

    if hasattr(polygon.boundary, "geoms"):
        boundaries = polygon.boundary.geoms
    else:
        boundaries = [polygon.boundary]

    # append interiors as single Line objects
    current = 0
    for boundary in boundaries:
        entities.append(Line(np.arange(len(boundary.coords)) + current))
        current += len(boundary.coords)
        # append the new vertex array
        vertices.append(np.array(boundary.coords))

    # make sure result arrays are numpy
    kwargs = {
        "entities": entities,
        "vertices": np.vstack(vertices) if len(vertices) > 0 else vertices,
    }

    return kwargs


def linestrings_to_path(multi) -> Dict:
    """
    Load shapely LineString objects into arguments to create a Path2D or Path3D.

    Parameters
    -------------
    multi : shapely.geometry.LineString or MultiLineString
      Input 2D or 3D geometry

    Returns
    -------------
    kwargs : Dict
      Keyword arguments for Path2D or Path3D constructor
    """
    import shapely

    # append to result as we go
    entities = []
    vertices = []

    if isinstance(multi, shapely.MultiLineString):
        multi = list(multi.geoms)
    else:
        multi = [multi]

    for line in multi:
        # only append geometry with points
        if hasattr(line, "coords"):
            coords = np.array(line.coords)
            if len(coords) < 2:
                continue
            entities.append(Line(np.arange(len(coords)) + len(vertices)))
            vertices.extend(coords)

    kwargs = {"entities": np.array(entities), "vertices": np.array(vertices)}
    return kwargs


def faces_to_path(mesh, face_ids=None, **kwargs):
    """
    Given a mesh and face indices find the outline edges and
    turn them into a Path3D.

    Parameters
    ------------
    mesh : trimesh.Trimesh
      Triangulated surface in 3D
    face_ids : (n,) int
      Indexes referencing mesh.faces

    Returns
    ---------
    kwargs : dict
      Kwargs for Path3D constructor
    """
    if face_ids is None:
        edges = mesh.edges_sorted
    else:
        # take advantage of edge ordering to index as single row
        edges = mesh.edges_sorted.reshape((-1, 6))[face_ids].reshape((-1, 2))
    # an edge which occurs onely once is on the boundary
    unique_edges = grouping.group_rows(edges, require_count=1)
    # add edges and vertices to kwargs
    kwargs.update(edges_to_path(edges=edges[unique_edges], vertices=mesh.vertices))

    return kwargs


def edges_to_path(edges: ArrayLike, vertices: ArrayLike, **kwargs) -> Dict:
    """
    Given an edge list of indices and associated vertices
    representing lines, generate kwargs for a Path object.

    Parameters
    -----------
    edges : (n, 2) int
      Vertex indices of line segments
    vertices : (m, dimension) float
      Vertex positions where dimension is 2 or 3

    Returns
    ----------
    kwargs : dict
      Kwargs for Path constructor
    """
    # sequence of ordered traversals
    dfs = graph.traversals(edges, mode="dfs")
    # make sure every consecutive index in DFS
    # traversal is an edge in the source edge list
    dfs_connected = graph.fill_traversals(dfs, edges=edges)
    # kwargs for Path constructor
    # turn traversals into Line objects
    lines = [Line(d) for d in dfs_connected]

    kwargs.update({"entities": lines, "vertices": vertices, "process": False})
    return kwargs



import os

from ... import util
from ...exchange import ply
from . import dxf, svg_io


def export_path(path, file_type=None, file_obj=None, **kwargs):
    """
    Export a Path object to a file- like object, or to a filename

    Parameters
    ---------
    file_obj:  None, str, or file object
      A filename string or a file-like object
    file_type: None or str
      File type, e.g.: 'svg', 'dxf'
    kwargs : passed to loader

    Returns
    ---------
    exported : str or bytes
      Data exported
    """
    # if file object is a string it is probably a file path
    # so we can split the extension to set the file type
    if isinstance(file_obj, str):
        file_type = util.split_extension(file_obj)

    # run the export
    export = _path_exporters[file_type](path, **kwargs)
    # if we've been passed files write the data
    _write_export(export=export, file_obj=file_obj)

    return export


def export_dict(path):
    """
    Export a path as a dict of kwargs for the Path constructor.
    """
    export_entities = [e.to_dict() for e in path.entities]
    export_object = {"entities": export_entities, "vertices": path.vertices.tolist()}
    return export_object


def _write_export(export, file_obj=None):
    """
    Write a string to a file.
    If file_obj isn't specified, return the string

    Parameters
    ---------
    export: a string of the export data
    file_obj: a file-like object or a filename
    """

    if file_obj is None:
        return export

    if hasattr(file_obj, "write"):
        out_file = file_obj
    else:
        # expand user and relative paths
        file_path = os.path.abspath(os.path.expanduser(file_obj))
        out_file = open(file_path, "wb")
    try:
        out_file.write(export)
    except TypeError:
        out_file.write(export.encode("utf-8"))

    out_file.close()

    return export


_path_exporters = {
    "dxf": dxf.export_dxf,
    "svg": svg_io.export_svg,
    "ply": ply.export_ply,
    "dict": export_dict,
}



import numpy as np


def test_resample_original():
    # check to see if `include_original` works

    from shapely.geometry import Polygon

    from trimesh.path.traversal import resample_path

    ori = np.array([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]], dtype=np.float64)

    re = resample_path(ori, step=0.25, include_original=True)

    a, b = Polygon(ori), Polygon(re)
    assert np.isclose(a.area, b.area)
    assert np.isclose(a.length, b.length)


if __name__ == "__main__":
    test_resample_original()



"""
registration.py
---------------

Functions for registering (aligning) point clouds with meshes.
"""

import numpy as np

from . import bounds, transformations, util
from .geometry import weighted_vertex_normals
from .points import PointCloud, plane_fit
from .transformations import transform_points
from .triangles import angles, cross, normals
from .typed import ArrayLike, Integer, Optional

try:
    import scipy.sparse as sparse
    from scipy.spatial import cKDTree
except BaseException as E:
    # wrapping just ImportError fails in some cases
    # will raise the error when someone tries to use KDtree
    from . import exceptions

    cKDTree = exceptions.ExceptionWrapper(E)
    sparse = exceptions.ExceptionWrapper(E)


def mesh_other(
    mesh,
    other,
    samples: Integer = 500,
    scale: bool = False,
    icp_first: Integer = 10,
    icp_final: Integer = 50,
    **kwargs,
):
    """
    Align a mesh with another mesh or a PointCloud using
    the principal axes of inertia as a starting point which
    is refined by iterative closest point.

    Parameters
    ------------
    mesh : trimesh.Trimesh object
      Mesh to align with other
    other : trimesh.Trimesh or (n, 3) float
      Mesh or points in space
    samples : int
      Number of samples from mesh surface to align
    scale : bool
      Allow scaling in transform
    icp_first : int
      How many ICP iterations for the 9 possible
      combinations of sign flippage
    icp_final : int
      How many ICP iterations for the closest
      candidate from the wider search
    kwargs : dict
      Passed through to `icp`, which passes through to `procrustes`

    Returns
    -----------
    mesh_to_other : (4, 4) float
      Transform to align mesh to the other object
    cost : float
      Average squared distance per point
    """

    def key_points(m, count):
        """
        Return a combination of mesh vertices and surface samples
        with vertices chosen by likelihood to be important
        to registration.
        """
        if len(m.vertices) < (count / 2):
            return np.vstack((m.vertices, m.sample(count - len(m.vertices))))
        else:
            return m.sample(count)

    if not util.is_instance_named(mesh, "Trimesh"):
        raise ValueError("mesh must be Trimesh object!")

    inverse = True
    search = mesh
    # if both are meshes use the smaller one for searching
    if util.is_instance_named(other, "Trimesh"):
        if len(mesh.vertices) > len(other.vertices):
            # do the expensive tree construction on the
            # smaller mesh and query the others points
            search = other
            inverse = False
            points = key_points(m=mesh, count=samples)
            points_mesh = mesh
        else:
            points_mesh = other
            points = key_points(m=other, count=samples)

        if points_mesh.is_volume:
            points_PIT = points_mesh.principal_inertia_transform
        else:
            points_PIT = points_mesh.bounding_box_oriented.principal_inertia_transform

    elif util.is_shape(other, (-1, 3)):
        # case where other is just points
        points = other
        points_PIT = bounds.oriented_bounds(points)[0]
    else:
        raise ValueError("other must be mesh or (n, 3) points!")

    # get the transform that aligns the search mesh principal
    # axes of inertia with the XYZ axis at the origin
    if search.is_volume:
        search_PIT = search.principal_inertia_transform
    else:
        search_PIT = search.bounding_box_oriented.principal_inertia_transform

    # transform that moves the principal axes of inertia
    # of the search mesh to be aligned with the best- guess
    # principal axes of the points
    search_to_points = np.dot(np.linalg.inv(points_PIT), search_PIT)

    # permutations of cube rotations
    # the principal inertia transform has arbitrary sign
    # along the 3 major axis so try all combinations of
    # 180 degree rotations with a quick first ICP pass
    cubes = np.array(
        [
            np.eye(4) * np.append(diag, 1)
            for diag in [
                [1, 1, 1],
                [1, 1, -1],
                [1, -1, 1],
                [-1, 1, 1],
                [-1, -1, 1],
                [-1, 1, -1],
                [1, -1, -1],
                [-1, -1, -1],
            ]
        ]
    )

    # loop through permutations and run iterative closest point
    costs = np.ones(len(cubes)) * np.inf
    transforms = [None] * len(cubes)
    centroid = search.centroid

    for i, flip in enumerate(cubes):
        # transform from points to search mesh
        # flipped around the centroid of search
        a_to_b = np.dot(
            transformations.transform_around(flip, centroid),
            np.linalg.inv(search_to_points),
        )

        # run first pass ICP
        matrix, _junk, cost = icp(
            a=points,
            b=search,
            initial=a_to_b,
            max_iterations=int(icp_first),
            scale=scale,
            **kwargs,
        )

        # save transform and costs from ICP
        transforms[i] = matrix
        costs[i] = cost

    # run a final ICP refinement step
    matrix, _junk, cost = icp(
        a=points,
        b=search,
        initial=transforms[np.argmin(costs)],
        max_iterations=int(icp_final),
        scale=scale,
        **kwargs,
    )

    # convert to per- point distance average
    cost /= len(points)

    # we picked the smaller mesh to construct the tree
    # on so we may have calculated a transform backwards
    # to save computation, so just invert matrix here
    if inverse:
        mesh_to_other = np.linalg.inv(matrix)
    else:
        mesh_to_other = matrix

    return mesh_to_other, cost


def procrustes(
    a: ArrayLike,
    b: ArrayLike,
    weights: Optional[ArrayLike] = None,
    reflection: bool = True,
    translation: bool = True,
    scale: bool = True,
    return_cost: bool = True,
):
    """
    Perform Procrustes' analysis to quickly align two corresponding
    point clouds subject to constraints. This is much cheaper than
    any other registration method but only applies if the two inputs
    correspond in order.

    Finds the transformation T mapping a to b which minimizes the
    square sum distances between Ta and b, also called the cost.

    Optionally filter the points in a and b via a binary weights array.
    Non-uniform weights are also supported, but won't yield the optimal rotation.

    Parameters
    ----------
    a : (n,3) float
      List of points in space
    b : (n,3) float
      List of points in space
    weights : (n,) float
      List of floats representing how much weight is assigned to each point.
      Binary entries can be used to filter the arrays; normalization is not required.
      Translation and scaling are adjusted according to the weighting.
      Note, however, that this method does not yield the optimal rotation for
      non-uniform weighting,
      as this would require an iterative, nonlinear optimization approach.
    reflection : bool
      If the transformation is allowed reflections
    translation : bool
      If the transformation is allowed translation and rotation.
    scale : bool
      If the transformation is allowed scaling
    return_cost : bool
      Whether to return the cost and transformed a as well

    Returns
    ----------
    matrix : (4,4) float
      The transformation matrix sending a to b
    transformed : (n,3) float
      The image of a under the transformation
    cost : float
      The cost of the transformation
    """

    a_original = np.asanyarray(a, dtype=np.float64)
    b_original = np.asanyarray(b, dtype=np.float64)
    if not util.is_shape(a_original, (-1, 3)) or not util.is_shape(b_original, (-1, 3)):
        raise ValueError("points must be (n,3)!")
    if len(a_original) != len(b_original):
        raise ValueError("a and b must contain same number of points!")
    # weights are set to uniform if not provided.
    if weights is None:
        weights = np.ones(len(a_original))
    w = np.maximum(np.asanyarray(weights, dtype=np.float64), 0)
    if len(w) != len(a):
        raise ValueError("weights must have same length as a and b!")
    w_norm = (w / w.sum()).reshape((-1, 1))

    # All zero entries are removed from further computations.
    # If weights is a binary array, the optimal solution can still be found by
    # simply removing the zero entries.
    nonzero_weights = w_norm[:, 0] > 0.0
    a_nonzero = a_original[nonzero_weights]
    b_nonzero = b_original[nonzero_weights]
    w_norm = w_norm[nonzero_weights]

    # Remove translation component
    if translation:
        # centers are (weighted) averages of the individual points.
        acenter = (a_nonzero * w_norm).sum(axis=0)
        bcenter = (b_nonzero * w_norm).sum(axis=0)
    else:
        acenter = np.zeros(a_nonzero.shape[1])
        bcenter = np.zeros(b_nonzero.shape[1])

    # Remove scale component
    if scale:
        # scale is the square root of the (weighted) average of the
        # squared difference between each point and the center.
        ascale = np.sqrt((((a_nonzero - acenter) ** 2) * w_norm).sum())
        bscale = np.sqrt((((b_nonzero - bcenter) ** 2) * w_norm).sum())
    else:
        ascale = 1
        bscale = 1

    # Use SVD to find optimal orthogonal matrix R
    # constrained to det(R) = 1 if necessary.

    target = np.dot(((b_nonzero - bcenter) / bscale).T, ((a_nonzero - acenter) / ascale))

    u, _s, vh = np.linalg.svd(target)

    if reflection:
        R = np.dot(u, vh)
    else:
        # no reflection allowed, so determinant must be 1.0
        R = np.dot(np.dot(u, np.diag([1, 1, np.linalg.det(np.dot(u, vh))])), vh)

    # Compute our 4D transformation matrix encoding
    # a -> (R @ (a - acenter)/ascale) * bscale + bcenter
    #    = (bscale/ascale)R @ a + (bcenter - (bscale/ascale)R @ acenter)
    translation = bcenter - (bscale / ascale) * np.dot(R, acenter)
    matrix = np.hstack((bscale / ascale * R, translation.reshape(-1, 1)))
    matrix = np.vstack((matrix, np.array([0.0] * (a.shape[1]) + [1.0]).reshape(1, -1)))

    if return_cost:
        # Transform the original input array, including zero-weighted points
        transformed = transform_points(a_original, matrix)
        # The cost is the (weighted) sum of the euclidean distances between
        # the transformed source points and the target points.
        cost = (((b_nonzero - transformed[nonzero_weights]) ** 2) * w_norm).sum()
        return matrix, transformed, cost

    return matrix


def icp(a, b, initial=None, threshold=1e-5, max_iterations=20, **kwargs):
    """
    Apply the iterative closest point algorithm to align a point cloud with
    another point cloud or mesh. Will only produce reasonable results if the
    initial transformation is roughly correct. Initial transformation can be
    found by applying Procrustes' analysis to a suitable set of landmark
    points (often picked manually).

    Parameters
    ----------
    a : (n,3) float
      List of points in space.
    b : (m,3) float or Trimesh
      List of points in space or mesh.
    initial : (4,4) float
      Initial transformation.
    threshold : float
      Stop when change in cost is less than threshold
    max_iterations : int
      Maximum number of iterations
    kwargs : dict
      Args to pass to procrustes

    Returns
    ----------
    matrix : (4,4) float
      The transformation matrix sending a to b
    transformed : (n,3) float
      The image of a under the transformation
    cost : float
      The cost of the transformation
    """

    a = np.asanyarray(a, dtype=np.float64)
    if not util.is_shape(a, (-1, 3)):
        raise ValueError("points must be (n,3)!")

    if initial is None:
        initial = np.eye(4)

    is_mesh = util.is_instance_named(b, "Trimesh")
    if not is_mesh:
        b = np.asanyarray(b, dtype=np.float64)
        if not util.is_shape(b, (-1, 3)):
            raise ValueError("points must be (n,3)!")
        btree = cKDTree(b)

    # transform a under initial_transformation
    a = transform_points(a, initial)
    total_matrix = initial

    # start with infinite cost
    old_cost = np.inf

    # avoid looping forever by capping iterations
    for _ in range(max_iterations):
        # Closest point in b to each point in a
        if is_mesh:
            closest, _distance, _faces = b.nearest.on_surface(a)
        else:
            _distances, ix = btree.query(a, 1)
            closest = b[ix]

        # align a with closest points
        matrix, transformed, cost = procrustes(a=a, b=closest, **kwargs)

        # update a with our new transformed points
        a = transformed
        total_matrix = np.dot(matrix, total_matrix)

        if old_cost - cost < threshold:
            break
        else:
            old_cost = cost

    return total_matrix, transformed, cost


def _normalize_by_source(source_mesh, target_geometry, target_positions):
    # Utility function to put the source mesh in [-1, 1]^3 and transform
    # target geometry accordingly
    if not util.is_instance_named(target_geometry, "Trimesh") and not isinstance(
        target_geometry, PointCloud
    ):
        vertices = np.asanyarray(target_geometry)
        target_geometry = PointCloud(vertices)
    centroid, scale = source_mesh.centroid, source_mesh.scale
    source_mesh.vertices = (source_mesh.vertices - centroid[None, :]) / scale
    # Dont forget to also transform the target positions
    target_geometry.vertices = (target_geometry.vertices - centroid[None, :]) / scale
    if target_positions is not None:
        target_positions = (target_positions - centroid[None, :]) / scale
    return target_geometry, target_positions, centroid, scale


def _denormalize_by_source(
    source_mesh, target_geometry, target_positions, result, centroid, scale
):
    # Utility function to transform source mesh from
    # [-1, 1]^3 to its original transform
    # and transform target geometry accordingly
    source_mesh.vertices = scale * source_mesh.vertices + centroid[None, :]
    target_geometry.vertices = scale * target_geometry.vertices + centroid[None, :]
    if target_positions is not None:
        target_positions = scale * target_positions + centroid[None, :]
    if isinstance(result, list):
        result = [scale * x + centroid[None, :] for x in result]
    else:
        result = scale * result + centroid[None, :]
    return result


def nricp_amberg(
    source_mesh,
    target_geometry,
    source_landmarks=None,
    target_positions=None,
    steps=None,
    eps=0.0001,
    gamma=1,
    distance_threshold=0.1,
    return_records=False,
    use_faces=True,
    use_vertex_normals=True,
    neighbors_count=8,
):
    """
    Non Rigid Iterative Closest Points

    Implementation of "Amberg et al. 2007: Optimal Step
    Nonrigid ICP Algorithms for Surface Registration."
    Allows to register non-rigidly a mesh on another or
    on a point cloud. The core algorithm is explained
    at the end of page 3 of the paper.

    Comparison between nricp_amberg and nricp_sumner:
    * nricp_amberg fits to the target mesh in less steps
    * nricp_amberg can generate sharp edges
      * only vertices and their neighbors are considered
    * nricp_sumner tend to preserve more the original shape
    * nricp_sumner parameters are easier to tune
    * nricp_sumner solves for triangle positions whereas
      nricp_amberg solves for vertex transforms
    * nricp_sumner is less optimized when wn > 0

    Parameters
    ----------
    source_mesh : Trimesh
        Source mesh containing both vertices and faces.
    target_geometry : Trimesh or PointCloud or (n, 3) float
        Target geometry. It can contain no faces or be a PointCloud.
    source_landmarks : (n,) int or ((n,) int, (n, 3) float)
        n landmarks on the the source mesh.
        Represented as vertex indices (n,) int.
        It can also be represented as a tuple of triangle
        indices and barycentric coordinates ((n,) int, (n, 3) float,).
    target_positions : (n, 3) float
        Target positions assigned to source landmarks
    steps : Core parameters of the algorithm
        Iterable of iterables (ws, wl, wn, max_iter,).
        ws is smoothness term, wl weights landmark importance, wn normal importance
        and max_iter is the maximum number of iterations per step.
    eps : float
        If the error decrease if inferior to this value, the current step ends.
    gamma : float
        Weight the translation part against the rotational/skew part.
        Recommended value : 1.
    distance_threshold : float
        Distance threshold to account for a vertex match or not.
    return_records : bool
        If True, also returns all the intermediate results. It can help debugging
        and tune the parameters to match a specific case.
    use_faces : bool
        If True and if target geometry has faces, use proximity.closest_point to find
        matching points. Else use scipy's cKDTree object.
    use_vertex_normals : bool
        If True and if target geometry has faces, interpolate the normals of the target
        geometry matching points.
        Else use face normals or estimated normals if target geometry has no faces.
    neighbors_count : int
        number of neighbors used for normal estimation. Only used if target geometry has
        no faces or if use_faces is False.

    Returns
    ----------
    result : (n, 3) float or List[(n, 3) float]
        The vertices positions of source_mesh such that it is registered non-rigidly
        onto the target geometry.
        If return_records is True, it returns the list of the vertex positions at each
        iteration.
    """

    def _solve_system(M_kron_G, D, vertices_weight, nearest, ws, nE, nV, Dl, Ul, wl):
        # Solve for Eq. 12
        U = nearest * vertices_weight[:, None]
        use_landmarks = Dl is not None and Ul is not None
        A_stack = [ws * M_kron_G, D.multiply(vertices_weight[:, None])]
        B_shape = (4 * nE + nV, 3)
        if use_landmarks:
            A_stack.append(wl * Dl)
            B_shape = (4 * nE + nV + Ul.shape[0], 3)
        A = sparse.csr_matrix(sparse.vstack(A_stack))
        B = sparse.lil_matrix(B_shape, dtype=np.float32)
        B[4 * nE : (4 * nE + nV), :] = U
        if use_landmarks:
            B[4 * nE + nV : (4 * nE + nV + Ul.shape[0]), :] = Ul * wl
        X = sparse.linalg.spsolve(A.T * A, A.T * B).toarray()
        return X

    def _node_arc_incidence(mesh, do_weight):
        # Computes node-arc incidence matrix of mesh (Eq.10)
        nV = mesh.edges.max() + 1
        nE = len(mesh.edges)
        rows = np.repeat(np.arange(nE), 2)
        cols = mesh.edges.flatten()
        data = np.ones(2 * nE, np.float32)
        data[1::2] = -1
        if do_weight:
            edge_lengths = np.linalg.norm(
                mesh.vertices[mesh.edges[:, 0]] - mesh.vertices[mesh.edges[:, 1]], axis=-1
            )
            data *= np.repeat(1 / edge_lengths, 2)
        return sparse.coo_matrix((data, (rows, cols)), shape=(nE, nV))

    def _create_D(vertex_3d_data):
        # Create Data matrix (Eq. 8)
        nV = len(vertex_3d_data)
        rows = np.repeat(np.arange(nV), 4)
        cols = np.arange(4 * nV)
        data = np.concatenate((vertex_3d_data, np.ones((nV, 1))), axis=-1).flatten()
        return sparse.csr_matrix((data, (rows, cols)), shape=(nV, 4 * nV))

    def _create_X(nV):
        # Create Unknowns Matrix (Eq. 1)
        X_ = np.concatenate((np.eye(3), np.array([[0, 0, 0]])), axis=0)
        return np.tile(X_, (nV, 1))

    def _create_Dl_Ul(D, source_mesh, source_landmarks, target_positions):
        # Create landmark terms (Eq. 11)
        Dl, Ul = None, None

        if source_landmarks is None or target_positions is None:
            # If no landmarks are provided, return None for both
            return Dl, Ul

        if isinstance(source_landmarks, tuple):
            source_tids, source_barys = source_landmarks
            source_tri_vids = source_mesh.faces[source_tids]
            # u * x1, v * x2 and w * x3 combined
            Dl = D[source_tri_vids.flatten(), :]
            Dl.data *= source_barys.flatten().repeat(np.diff(Dl.indptr))
            x0 = source_mesh.vertices[source_tri_vids[:, 0]]
            x1 = source_mesh.vertices[source_tri_vids[:, 1]]
            x2 = source_mesh.vertices[source_tri_vids[:, 2]]
            Ul0 = (
                target_positions
                - x1 * source_barys[:, 1, None]
                - x2 * source_barys[:, 2, None]
            )
            Ul1 = (
                target_positions
                - x0 * source_barys[:, 0, None]
                - x2 * source_barys[:, 2, None]
            )
            Ul2 = (
                target_positions
                - x0 * source_barys[:, 0, None]
                - x1 * source_barys[:, 1, None]
            )
            Ul = np.zeros((Ul0.shape[0] * 3, 3))
            Ul[0::3] = Ul0  # y - v * x2 + w * x3
            Ul[1::3] = Ul1  # y - u * x1 + w * x3
            Ul[2::3] = Ul2  # y - u * x1 + v * x2
        else:
            Dl = D[source_landmarks, :]
            Ul = target_positions
        return Dl, Ul

    target_geometry, target_positions, centroid, scale = _normalize_by_source(
        source_mesh, target_geometry, target_positions
    )

    # Number of edges and vertices in source mesh
    nE = len(source_mesh.edges)
    nV = len(source_mesh.vertices)

    # Initialize transformed vertices
    transformed_vertices = source_mesh.vertices.copy()
    # Node-arc incidence (M in Eq. 10)
    M = _node_arc_incidence(source_mesh, True)
    # G (Eq. 10)
    G = np.diag([1, 1, 1, gamma])
    # M kronecker G (Eq. 10)
    M_kron_G = sparse.kron(M, G)
    # D (Eq. 8)
    D = _create_D(source_mesh.vertices)
    # D but for normal computation from the transformations X
    DN = _create_D(source_mesh.vertex_normals)
    # Unknowns 4x3 transformations X (Eq. 1)
    X = _create_X(nV)
    # Landmark related terms (Eq. 11)
    Dl, Ul = _create_Dl_Ul(D, source_mesh, source_landmarks, target_positions)

    # Parameters of the algorithm (Eq. 6)
    # order : Alpha, Beta, normal weighting, and max iteration for step
    if steps is None:
        steps = [
            [0.01, 10, 0.5, 10],
            [0.02, 5, 0.5, 10],
            [0.03, 2.5, 0.5, 10],
            [0.01, 0, 0.0, 10],
        ]
    if return_records:
        records = [transformed_vertices]

    # Main loop
    for ws, wl, wn, max_iter in steps:
        # If normals are estimated from points and if there are less
        # than 3 points per query, avoid normal estimation
        if not use_faces and neighbors_count < 3:
            wn = 0

        last_error = np.finfo(np.float32).max
        error = np.finfo(np.float16).max
        cpt_iter = 0

        # Current step iterations loop
        while last_error - error > eps and (max_iter is None or cpt_iter < max_iter):
            qres = _from_mesh(
                target_geometry,
                transformed_vertices,
                from_vertices_only=not use_faces,
                return_normals=wn > 0,
                return_interpolated_normals=wn > 0 and use_vertex_normals,
                neighbors_count=neighbors_count,
            )

            # Data weighting
            vertices_weight = np.ones(nV)
            vertices_weight[qres["distances"] > distance_threshold] = 0

            if wn > 0 and "normals" in qres:
                target_normals = qres["normals"]
                if use_vertex_normals and "interpolated_normals" in qres:
                    target_normals = qres["interpolated_normals"]
                # Normal weighting = multiplying weights by cosines^wn
                source_normals = DN * X
                dot = util.diagonal_dot(source_normals, target_normals)
                # Normal orientation is only known for meshes as target
                dot = np.clip(dot, 0, 1) if use_faces else np.abs(dot)
                vertices_weight = vertices_weight * dot**wn

            # Actual system solve
            X = _solve_system(
                M_kron_G, D, vertices_weight, qres["nearest"], ws, nE, nV, Dl, Ul, wl
            )
            transformed_vertices = D * X
            last_error = error
            error_vec = np.linalg.norm(qres["nearest"] - transformed_vertices, axis=-1)
            error = (error_vec * vertices_weight).mean()
            if return_records:
                records.append(transformed_vertices)
            cpt_iter += 1

    if return_records:
        result = records
    else:
        result = transformed_vertices

    result = _denormalize_by_source(
        source_mesh, target_geometry, target_positions, result, centroid, scale
    )
    return result


def _from_mesh(
    mesh,
    input_points,
    from_vertices_only=False,
    return_barycentric_coordinates=False,
    return_normals=False,
    return_interpolated_normals=False,
    neighbors_count=10,
    **kwargs,
):
    """
    Find the the closest points and associated attributes from a Trimesh.

    Parameters
    -----------
    mesh : Trimesh
        Trimesh from which the query is performed
    input_points : (m, 3) float
        Input query points
    from_vertices_only : bool
        If True, consider only the vertices and not the faces
    return_barycentric_coordinates : bool
        If True, return the barycentric coordinates
    return_normals : bool
        If True, compute the normals at each closest point
    return_interpolated_normals : bool
        If True, return the interpolated normal at each closest point
    neighbors_count : int
        The number of closest neighbors to query
    kwargs : dict
        Dict to accept other key word arguments (not used)
    Returns
    ----------
    qres : Dict
      Dictionary containing :
       - nearest points (m, 3) with key 'nearest'
       - distances to nearest point (m,) with key 'distances'
       - support triangle indices of the nearest points (m,) with key 'tids'
       - [optional] normals at nearest points (m,3) with key 'normals'
       - [optional] barycentric coordinates in support triangles (m,3) with key
         'barycentric_coordinates'
       - [optional] interpolated normals (m,3) with key 'interpolated_normals'
    """
    input_points = np.asanyarray(input_points)
    neighbors_count = min(neighbors_count, len(mesh.vertices))

    if from_vertices_only or len(mesh.faces) == 0:
        # Consider only the vertices
        return _from_points(
            mesh.vertices,
            input_points,
            mesh.kdtree,
            return_normals=return_normals,
            neighbors_count=neighbors_count,
        )
    # Else if we consider faces, use proximity.closest_point
    qres = {}
    from .proximity import closest_point
    from .triangles import points_to_barycentric

    qres["nearest"], qres["distances"], qres["tids"] = closest_point(mesh, input_points)

    if return_normals:
        qres["normals"] = mesh.face_normals[qres["tids"]]
    if return_barycentric_coordinates or return_interpolated_normals:
        qres["barycentric_coordinates"] = points_to_barycentric(
            mesh.vertices[mesh.faces[qres["tids"]]], qres["nearest"]
        )

        if return_interpolated_normals:
            # Interpolation from barycentric coordinates
            qres["interpolated_normals"] = np.einsum(
                "ij,ijk->ik",
                qres["barycentric_coordinates"],
                mesh.vertex_normals[mesh.faces[qres["tids"]]],
            )
    return qres


def _from_points(
    target_points,
    input_points,
    kdtree=None,
    return_normals=False,
    neighbors_count=10,
    **kwargs,
):
    """
    Find the the closest points and associated attributes
    from a set of 3D points.

    Parameters
    -----------
    target_points : (n, 3) float
      Points from which the query is performed
    input_points : (m, 3) float
      Input query points
    kdtree : scipy.cKDTree
      KDTree used for query. Computed if not provided
    return_normals : bool
      If True, compute the normals at each nearest point
    neighbors_count : int
      The number of closest neighbors to query
    kwargs : dict
      Dict to accept other key word arguments (not used)

    Returns
    ----------
    qres : Dict
      Dictionary containing :
       - nearest points (m, 3) with key 'nearest'
       - distances to nearest point (m,) with key 'distances'
       - vertex indices of the nearest points (m,) with key 'vertex_indices'
       - [optional] normals at nearest points (m,3) with key 'normals'
    """
    # Empty result
    target_points = np.asanyarray(target_points)
    input_points = np.asanyarray(input_points)
    neighbors_count = min(neighbors_count, len(target_points))
    qres = {}
    if kdtree is None:
        kdtree = cKDTree(target_points)

    if return_normals:
        assert neighbors_count >= 3
        distances, indices = kdtree.query(input_points, k=neighbors_count)
        nearest = target_points[indices, :]
        qres["normals"] = plane_fit(nearest)[1]
        qres["nearest"] = nearest[:, 0]
        qres["distances"] = distances[:, 0]
        qres["vertex_indices"] = indices[:, 0]
    else:
        qres["distances"], qres["vertex_indices"] = kdtree.query(input_points)
        qres["nearest"] = target_points[qres["vertex_indices"], :]

    return qres


def nricp_sumner(
    source_mesh,
    target_geometry,
    source_landmarks=None,
    target_positions=None,
    steps=None,
    distance_threshold=0.1,
    return_records=False,
    use_faces=True,
    use_vertex_normals=True,
    neighbors_count=8,
    face_pairs_type="vertex",
):
    """
    Non Rigid Iterative Closest Points

    Implementation of the correspondence computation part of
    "Sumner and Popovic 2004: Deformation Transfer for Triangle Meshes"
    Allows to register non-rigidly a mesh on another geometry.

    Comparison between nricp_amberg and nricp_sumner:
    * nricp_amberg fits to the target mesh in less steps
    * nricp_amberg can generate sharp edges (only vertices and their
        neighbors are considered)
    * nricp_sumner tend to preserve more the original shape
    * nricp_sumner parameters are easier to tune
    * nricp_sumner solves for triangle positions whereas nricp_amberg solves for
        vertex transforms
    * nricp_sumner is less optimized when wn > 0

    Parameters
    ----------
    source_mesh : Trimesh
        Source mesh containing both vertices and faces.
    target_geometry : Trimesh or PointCloud or (n, 3) float
        Target geometry. It can contain no faces or be a PointCloud.
    source_landmarks : (n,) int or ((n,) int, (n, 3) float)
        n landmarks on the the source mesh.
        Represented as vertex indices (n,) int.
        It can also be represented as a tuple of triangle indices and barycentric
        coordinates ((n,) int, (n, 3) float,).
    target_positions : (n, 3) float
        Target positions assigned to source landmarks
    steps : Core parameters of the algorithm
        Iterable of iterables (wc, wi, ws, wl, wn).
        wc is the correspondence term (strength of fitting), wi is the identity term
        (recommended value : 0.001), ws is smoothness term, wl weights the landmark
        importance and wn the normal importance.
    distance_threshold : float
        Distance threshold to account for a vertex match or not.
    return_records : bool
        If True, also returns all the intermediate results. It can help debugging
        and tune the parameters to match a specific case.
    use_faces : bool
        If True and if target geometry has faces, use proximity.closest_point to find
        matching points. Else use scipy's cKDTree object.
    use_vertex_normals : bool
        If True and if target geometry has faces, interpolate the normals of the target
        geometry matching points.
        Else use face normals or estimated normals if target geometry has no faces.
    neighbors_count : int
        number of neighbors used for normal estimation. Only used if target geometry has
        no faces or if use_faces is False.
    face_pairs_type : str 'vertex' or 'edge'
        Method to determine face pairs used in the smoothness cost. 'vertex' yields
        smoother results.


    Returns
    ----------
    result : (n, 3) float or List[(n, 3) float]
        The vertices positions of source_mesh such that it is registered non-rigidly
        onto the target geometry.
        If return_records is True, it returns the list of the vertex positions at each
        iteration.
    """

    def _construct_transform_matrix(faces, Vinv, size):
        # Utility function for constructing the per-frame transforms
        _construct_transform_matrix._row = np.array([0, 1, 2] * 4)
        nV = len(Vinv)
        rows = np.tile(_construct_transform_matrix._row, nV) + 3 * np.repeat(
            np.arange(nV), 12
        )
        cols = np.repeat(faces.flat, 3)
        minus_inv_sum = -Vinv.sum(axis=1)
        Vinv_flat = Vinv.reshape(nV, 9)
        data = np.concatenate((minus_inv_sum, Vinv_flat), axis=-1).flatten()
        return sparse.coo_matrix((data, (rows, cols)), shape=(3 * nV, size), dtype=float)

    def _build_tetrahedrons(mesh):
        # UUtility function for constructing the frames
        v4_vec = mesh.face_normals
        v1 = mesh.triangles[:, 0]
        v2 = mesh.triangles[:, 1]
        v3 = mesh.triangles[:, 2]
        v4 = v1 + v4_vec
        vertices = np.concatenate((mesh.vertices, v4))
        nV, nT = len(mesh.vertices), len(mesh.faces)
        v4_indices = np.arange(nV, nV + nT)[:, None]
        tetrahedrons = np.concatenate((mesh.faces, v4_indices), axis=-1)
        frames = np.concatenate(
            ((v2 - v1)[..., None], (v3 - v1)[..., None], v4_vec[..., None]), axis=-1
        )
        return vertices, tetrahedrons, frames

    def _construct_identity_cost(vtet, tet, Vinv):
        # Utility function for constructing the identity cost
        AEi = _construct_transform_matrix(
            tet,
            Vinv,
            len(vtet),
        ).tocsr()
        Bi = np.tile(np.identity(3, dtype=float), (len(tet), 1))
        return AEi, Bi

    def _construct_smoothness_cost(vtet, tet, Vinv, face_pairs):
        # Utility function for constructing the smoothness (stiffness) cost
        AEs_r = _construct_transform_matrix(
            tet[face_pairs[:, 0]], Vinv[face_pairs[:, 0]], len(vtet)
        ).tocsr()
        AEs_l = _construct_transform_matrix(
            tet[face_pairs[:, 1]], Vinv[face_pairs[:, 1]], len(vtet)
        ).tocsr()
        AEs = (AEs_r - AEs_l).tocsc()
        AEs.eliminate_zeros()
        Bs = np.zeros((len(face_pairs) * 3, 3))
        return AEs, Bs

    def _construct_landmark_cost(vtet, source_mesh, source_landmarks):
        # Utility function for constructing the landmark cost
        if source_landmarks is None:
            return None, np.ones(len(source_mesh.vertices), dtype=bool)
        if isinstance(source_landmarks, tuple):
            # If the input source landmarks are in barycentric form
            source_landmarks_tids, source_landmarks_barys = source_landmarks
            source_landmarks_vids = source_mesh.faces[source_landmarks_tids]
            nL, nVT = len(source_landmarks_tids), len(vtet)

            rows = np.repeat(np.arange(nL), 3)
            cols = source_landmarks_vids.flat
            data = source_landmarks_barys.flat

            AEl = sparse.coo_matrix((data, (rows, cols)), shape=(nL, nVT))
            marker_vids = source_landmarks_vids[
                source_landmarks_barys > np.finfo(np.float16).eps
            ]
            non_markers_mask = np.ones(len(source_mesh.vertices), dtype=bool)
            non_markers_mask[marker_vids] = False
        else:
            # Else if they are in vertex index form
            nL, nVT = len(source_landmarks), len(vtet)
            rows = np.arange(nL)
            cols = source_landmarks.flat
            data = np.ones(nL)
            AEl = sparse.coo_matrix((data, (rows, cols)), shape=(nL, nVT))
            non_markers_mask = np.ones(len(source_mesh.vertices), dtype=bool)
            non_markers_mask[source_landmarks] = False
        return AEl, non_markers_mask

    def _construct_correspondence_cost(points, non_markers_mask, size):
        # Utility function for constructing the correspondence cost
        AEc = sparse.identity(size, dtype=float, format="csc")[: len(non_markers_mask)]
        AEc = AEc[non_markers_mask]
        Bc = points[non_markers_mask]
        return AEc, Bc

    def _compute_vertex_normals(vertices, faces):
        # Utility function for computing source vertex normals
        mesh_triangles = vertices[faces]
        mesh_triangles_cross = cross(mesh_triangles)
        mesh_face_normals = normals(
            triangles=mesh_triangles, crosses=mesh_triangles_cross
        )[0]
        mesh_face_angles = angles(mesh_triangles)
        mesh_normals = weighted_vertex_normals(
            vertex_count=nV,
            faces=faces,
            face_normals=mesh_face_normals,
            face_angles=mesh_face_angles,
        )
        return mesh_normals

    # First, normalize the source and target to [-1, 1]^3
    (target_geometry, target_positions, centroid, scale) = _normalize_by_source(
        source_mesh, target_geometry, target_positions
    )
    nV = len(source_mesh.vertices)
    use_landmarks = source_landmarks is not None and target_positions is not None

    if steps is None:
        steps = [
            # [wc, wi, ws, wl, wn],
            [1, 0.001, 1.0, 1000, 0],
            [1, 0.001, 1.0, 1000, 0],
            [10, 0.001, 1.0, 1000, 0],
            [100, 0.001, 1.0, 1000, 0],
        ]

    source_vtet, source_tet, V = _build_tetrahedrons(source_mesh)
    Vinv = np.linalg.inv(V)

    # List of (n, 2) faces index which share a vertex
    if face_pairs_type == "vertex":
        face_pairs = source_mesh.face_neighborhood
    else:
        face_pairs = source_mesh.face_adjacency

    # Construct the cost matrices
    # Identity cost (Eq. 12)
    AEi, Bi = _construct_identity_cost(source_vtet, source_tet, Vinv)
    # Smoothness cost (Eq. 11)
    AEs, Bs = _construct_smoothness_cost(source_vtet, source_tet, Vinv, face_pairs)
    # Landmark cost (Eq. 13)
    AEl, non_markers_mask = _construct_landmark_cost(
        source_vtet, source_mesh, source_landmarks
    )

    transformed_vertices = source_vtet.copy()
    if return_records:
        records = [transformed_vertices[:nV]]

    # Main loop
    for i, (wc, wi, ws, wl, wn) in enumerate(steps):
        Astack = [AEi * wi, AEs * ws]
        Bstack = [Bi * wi, Bs * ws]

        if use_landmarks:
            Astack.append(AEl * wl)
            Bstack.append(target_positions * wl)

        if (i > 0 or not use_landmarks) and wc > 0:
            # Query the nearest points
            qres = _from_mesh(
                target_geometry,
                transformed_vertices[:nV],
                from_vertices_only=not use_faces,
                return_normals=wn > 0,
                return_interpolated_normals=(use_vertex_normals and wn > 0),
                neighbors_count=neighbors_count,
            )

            # Correspondence cost (Eq. 13)
            AEc, Bc = _construct_correspondence_cost(
                qres["nearest"], non_markers_mask, len(source_vtet)
            )
            vertices_weight = np.ones(nV)
            vertices_weight[qres["distances"] > distance_threshold] = 0
            if wn > 0 or "normals" in qres:
                target_normals = qres["normals"]
                if use_vertex_normals and "interpolated_normals" in qres:
                    target_normals = qres["interpolated_normals"]
                # Normal weighting : multiplying weights by cosines^wn
                source_normals = _compute_vertex_normals(
                    transformed_vertices, source_mesh.faces
                )
                dot = util.diagonal_dot(source_normals, target_normals)
                # Normal orientation is only known for meshes as target
                dot = np.clip(dot, 0, 1) if use_faces else np.abs(dot)
                vertices_weight = vertices_weight * dot**wn

            # Account for vertices' weight
            AEc.data *= vertices_weight[non_markers_mask][AEc.indices]
            Bc *= vertices_weight[non_markers_mask, None]

            Astack.append(AEc * wc)
            Bstack.append(Bc * wc)

        # Now solve Eq. 14 ...
        A = sparse.vstack(Astack, format="csc")
        A.eliminate_zeros()
        b = np.concatenate(Bstack)

        LU = sparse.linalg.splu((A.T * A).tocsc())
        transformed_vertices = LU.solve(A.T * b)
        # done !

        if return_records:
            records.append(transformed_vertices[:nV])

    if return_records:
        result = records
    else:
        result = transformed_vertices[:nV]

    result = _denormalize_by_source(
        source_mesh, target_geometry, target_positions, result, centroid, scale
    )
    return result



"""
base.py
-------------

The base class for `Visual` objects
"""

import abc

from ..util import ABC


class Visuals(ABC):
    """
    Parent of Visual classes.
    """

    @property
    @abc.abstractmethod
    def kind(self):
        pass

    @abc.abstractmethod
    def update_vertices(self, mask):
        pass

    @abc.abstractmethod
    def update_faces(self, mask):
        pass

    @abc.abstractmethod
    def concatenate(self, other):
        pass

    @abc.abstractmethod
    def __hash__(self):
        pass

    @abc.abstractmethod
    def copy(self):
        pass

    def __add__(self, other):
        """
        Concatenate two ColorVisuals objects into a single object.

        Parameters
        -----------
        other : Visuals
          Other visual to concatenate

        Returns
        -----------
        result : Visuals
          Object containing information from current
          object and other in the order (self, other)
        """
        return self.concatenate(other)



import copy
import io
import uuid

import numpy as np

from .. import util, visual
from ..constants import log
from ..util import unique_name

_EYE = np.eye(4)
_EYE.flags.writeable = False


def load_collada(file_obj, resolver=None, ignore_broken=True, **kwargs):
    """
    Load a COLLADA (.dae) file into a list of trimesh kwargs.

    Parameters
    ----------
    file_obj : file object
      Containing a COLLADA file
    resolver : trimesh.visual.Resolver or None
      For loading referenced files, like texture images
    ignore_broken: bool
      Ignores broken references during loading:
        [collada.common.DaeUnsupportedError,
         collada.common.DaeBrokenRefError]
    kwargs : **
      Passed to trimesh.Trimesh.__init__

    Returns
    -------
    loaded : list of dict
      kwargs for Trimesh constructor
    """
    import collada

    if ignore_broken:
        ignores = [
            collada.common.DaeError,
            collada.common.DaeIncompleteError,
            collada.common.DaeMalformedError,
            collada.common.DaeBrokenRefError,
            collada.common.DaeUnsupportedError,
            collada.common.DaeIncompleteError,
        ]
    else:
        ignores = None

    # load scene using pycollada
    c = collada.Collada(file_obj, ignore=ignores)

    # Create material map from Material ID to trimesh material
    material_map = {}
    for m in c.materials:
        effect = m.effect
        material_map[m.id] = _parse_material(effect, resolver)

    unit = c.assetInfo.unitmeter
    if unit is None or np.isclose(unit, 1.0):
        metadata = {"units": "meters"}
    else:
        metadata = {"units": f"{unit} * meters"}

    # name : kwargs
    meshes = {}
    # increments to enable `unique_name` to avoid n^2 behavior
    meshes_count = {}
    # list of dict
    graph = []

    for node in c.scene.nodes:
        _parse_node(
            node=node,
            parent_matrix=_EYE,
            material_map=material_map,
            meshes=meshes,
            meshes_count=meshes_count,
            graph=graph,
            resolver=resolver,
            metadata=metadata,
        )

    return {"class": "Scene", "graph": graph, "geometry": meshes}


def export_collada(mesh, **kwargs):
    """
    Export a mesh or a list of meshes as a COLLADA .dae file.

    Parameters
    -----------
    mesh: Trimesh object or list of Trimesh objects
        The mesh(es) to export.

    Returns
    -----------
    export: str, string of COLLADA format output
    """
    import collada

    meshes = mesh
    if not isinstance(mesh, (list, tuple, set, np.ndarray)):
        meshes = [mesh]

    c = collada.Collada()
    nodes = []
    for i, m in enumerate(meshes):
        # Load uv, colors, materials
        uv = None
        colors = None
        mat = _unparse_material(None)
        if m.visual.defined:
            if m.visual.kind == "texture":
                mat = _unparse_material(m.visual.material)
                uv = m.visual.uv
            elif m.visual.kind == "vertex":
                colors = (m.visual.vertex_colors / 255.0)[:, :3]
                mat.effect.diffuse = np.array(m.visual.main_color) / 255.0
            elif m.visual.kind == "face":
                mat.effect.diffuse = np.array(m.visual.main_color) / 255.0
        c.effects.append(mat.effect)
        c.materials.append(mat)

        # Create geometry object
        vertices = collada.source.FloatSource(
            "verts-array", m.vertices.flatten(), ("X", "Y", "Z")
        )
        normals = collada.source.FloatSource(
            "normals-array", m.vertex_normals.flatten(), ("X", "Y", "Z")
        )
        input_list = collada.source.InputList()
        input_list.addInput(0, "VERTEX", "#verts-array")
        input_list.addInput(1, "NORMAL", "#normals-array")
        arrays = [vertices, normals]
        if (uv is not None) and (len(uv) > 0):
            texcoords = collada.source.FloatSource(
                "texcoords-array", uv.flatten(), ("U", "V")
            )
            input_list.addInput(2, "TEXCOORD", "#texcoords-array")
            arrays.append(texcoords)
        if colors is not None:
            idx = 2
            if uv:
                idx = 3
            colors = collada.source.FloatSource(
                "colors-array", colors.flatten(), ("R", "G", "B")
            )
            input_list.addInput(idx, "COLOR", "#colors-array")
            arrays.append(colors)
        geom = collada.geometry.Geometry(c, uuid.uuid4().hex, uuid.uuid4().hex, arrays)
        indices = np.repeat(m.faces.flatten(), len(arrays))

        matref = f"material{i}"
        triset = geom.createTriangleSet(indices, input_list, matref)
        geom.primitives.append(triset)
        c.geometries.append(geom)

        matnode = collada.scene.MaterialNode(matref, mat, inputs=[])
        geomnode = collada.scene.GeometryNode(geom, [matnode])
        node = collada.scene.Node(f"node{i}", children=[geomnode])
        nodes.append(node)
    scene = collada.scene.Scene("scene", nodes)
    c.scenes.append(scene)
    c.scene = scene

    b = io.BytesIO()
    c.write(b)
    b.seek(0)
    return b.read()


def _parse_node(
    node, parent_matrix, material_map, meshes, meshes_count, graph, resolver, metadata
):
    """
    Recursively parse COLLADA scene nodes.
    """
    import collada

    # Parse mesh node
    if isinstance(node, collada.scene.GeometryNode):
        geometry = node.geometry

        # Create local material map from material symbol to actual material
        local_material_map = {}
        for mn in node.materials:
            symbol = mn.symbol
            m = mn.target
            if m.id in material_map:
                local_material_map[symbol] = material_map[m.id]
            else:
                local_material_map[symbol] = _parse_material(m, resolver)

        # Iterate over primitives of geometry
        for primitive in geometry.primitives:
            if isinstance(primitive, collada.polylist.Polylist):
                primitive = primitive.triangleset()
            if isinstance(primitive, collada.triangleset.TriangleSet):
                vertex = primitive.vertex
                if vertex is None:
                    continue
                vertex_index = primitive.vertex_index
                vertices = vertex[vertex_index].reshape(len(vertex_index) * 3, 3)

                # Get normals if present
                normals = None
                if primitive.normal is not None:
                    normal = primitive.normal
                    normal_index = primitive.normal_index
                    normals = normal[normal_index].reshape(len(normal_index) * 3, 3)

                # Get colors if present
                colors = None
                s = primitive.sources
                if "COLOR" in s and len(s["COLOR"]) > 0 and len(primitive.index) > 0:
                    color = s["COLOR"][0][4].data
                    color_index = primitive.index[:, :, s["COLOR"][0][0]]
                    colors = color[color_index].reshape(len(color_index) * 3, -1)

                faces = np.arange(vertices.shape[0]).reshape(vertices.shape[0] // 3, 3)

                # Get UV coordinates if possible
                vis = None
                if primitive.material in local_material_map:
                    material = copy.copy(local_material_map[primitive.material])
                    uv = None
                    if len(primitive.texcoordset) > 0:
                        texcoord = primitive.texcoordset[0]
                        texcoord_index = primitive.texcoord_indexset[0]
                        uv = texcoord[texcoord_index].reshape(
                            (len(texcoord_index) * 3, 2)
                        )
                    vis = visual.texture.TextureVisuals(uv=uv, material=material)

                geom_name = unique_name(geometry.id, contains=meshes, counts=meshes_count)
                meshes[geom_name] = {
                    "vertices": vertices,
                    "faces": faces,
                    "vertex_normals": normals,
                    "vertex_colors": colors,
                    "visual": vis,
                    "metadata": metadata,
                }

                graph.append(
                    {
                        "frame_to": geom_name,
                        "matrix": parent_matrix,
                        "geometry": geom_name,
                    }
                )

    # recurse down tree for nodes with children
    elif isinstance(node, collada.scene.Node):
        if node.children is not None:
            for child in node.children:
                # create the new matrix
                matrix = np.dot(parent_matrix, node.matrix)
                # parse the child node
                _parse_node(
                    node=child,
                    parent_matrix=matrix,
                    material_map=material_map,
                    meshes=meshes,
                    meshes_count=meshes_count,
                    graph=graph,
                    resolver=resolver,
                    metadata=metadata,
                )

    elif isinstance(node, collada.scene.CameraNode):
        # TODO: convert collada cameras to trimesh cameras
        pass
    elif isinstance(node, collada.scene.LightNode):
        # TODO: convert collada lights to trimesh lights
        pass


def _load_texture(file_name, resolver):
    """
    Load a texture from a file into a PIL image.
    """
    from PIL import Image

    file_data = resolver.get(file_name)
    image = Image.open(util.wrap_as_stream(file_data))
    return image


def _parse_material(effect, resolver):
    """
    Turn a COLLADA effect into a trimesh material.
    """
    import collada

    # Compute base color
    baseColorFactor = np.ones(4)
    baseColorTexture = None
    if isinstance(effect.diffuse, collada.material.Map):
        try:
            baseColorTexture = _load_texture(
                effect.diffuse.sampler.surface.image.path, resolver
            )
        except BaseException:
            log.debug("unable to load base texture", exc_info=True)
    elif effect.diffuse is not None:
        baseColorFactor = effect.diffuse

    # Compute emission color
    emissiveFactor = np.zeros(3)
    emissiveTexture = None
    if isinstance(effect.emission, collada.material.Map):
        try:
            emissiveTexture = _load_texture(
                effect.diffuse.sampler.surface.image.path, resolver
            )
        except BaseException:
            log.warning("unable to load emissive texture", exc_info=True)
    elif effect.emission is not None:
        emissiveFactor = effect.emission[:3]

    # Compute roughness
    roughnessFactor = 1.0
    if (
        not isinstance(effect.shininess, collada.material.Map)
        and effect.shininess is not None
    ):
        try:
            shininess_value = float(effect.shininess)
            roughnessFactor = np.sqrt(2.0 / (2.0 + shininess_value))
        except (TypeError, ValueError):
            log.warning(
                f"Invalid shininess value: {effect.shininess}, using default roughness"
            )

    # Compute metallic factor
    metallicFactor = 0.0

    # Compute normal texture
    normalTexture = None
    if effect.bumpmap is not None:
        try:
            normalTexture = _load_texture(
                effect.bumpmap.sampler.surface.image.path, resolver
            )
        except BaseException:
            log.warning("unable to load bumpmap", exc_info=True)

    # Compute opacity
    if effect.transparent is not None and not isinstance(
        effect.transparent, collada.material.Map
    ):
        baseColorFactor = tuple(
            np.append(baseColorFactor[:3], float(effect.transparent[3]))
        )

    return visual.material.PBRMaterial(
        emissiveFactor=emissiveFactor,
        emissiveTexture=emissiveTexture,
        normalTexture=normalTexture,
        baseColorTexture=baseColorTexture,
        baseColorFactor=baseColorFactor,
        metallicFactor=metallicFactor,
        roughnessFactor=roughnessFactor,
    )


def _unparse_material(material):
    """
    Turn a trimesh material into a COLLADA material.
    """
    import collada

    # TODO EXPORT TEXTURES
    if isinstance(material, visual.material.PBRMaterial):
        diffuse = material.baseColorFactor
        if diffuse is None:
            diffuse = np.array([255.0, 255.0, 255.0, 255.0])
        diffuse = diffuse / 255.0
        if diffuse is not None:
            diffuse = list(diffuse)

        emission = material.emissiveFactor
        if emission is not None:
            emission = [float(emission[0]), float(emission[1]), float(emission[2]), 1.0]

        shininess = material.roughnessFactor
        if shininess is None:
            shininess = 1.0
        if shininess is not None:
            shininess = 2.0 / shininess**2 - 2.0

        effect = collada.material.Effect(
            uuid.uuid4().hex,
            params=[],
            shadingtype="phong",
            diffuse=diffuse,
            emission=emission,
            specular=[1.0, 1.0, 1.0, 1.0],
            shininess=float(shininess),
        )
        material = collada.material.Material(uuid.uuid4().hex, "pbrmaterial", effect)
    else:
        effect = collada.material.Effect(uuid.uuid4().hex, params=[], shadingtype="phong")
        material = collada.material.Material(uuid.uuid4().hex, "defaultmaterial", effect)
    return material


def load_zae(file_obj, resolver=None, **kwargs):
    """
    Load a ZAE file, which is just a zipped DAE file.

    Parameters
    -------------
    file_obj : file object
      Contains ZAE data
    resolver : trimesh.visual.Resolver
      Resolver to load additional assets
    kwargs : dict
      Passed to load_collada

    Returns
    ------------
    loaded : dict
      Results of loading
    """

    # a dict, {file name : file object}
    archive = util.decompress(file_obj, file_type="zip")

    # load the first file with a .dae extension
    file_name = next(i for i in archive.keys() if i.lower().endswith(".dae"))

    # a resolver so the loader can load textures / etc
    resolver = visual.resolvers.ZipResolver(archive)

    # run the regular collada loader
    loaded = load_collada(archive[file_name], resolver=resolver, **kwargs)
    return loaded


# only provide loaders if `pycollada` is installed
_collada_loaders = {}
_collada_exporters = {}
if util.has_module("collada"):
    _collada_loaders["dae"] = load_collada
    _collada_loaders["zae"] = load_zae
    _collada_exporters["dae"] = export_collada
else:
    # store an exception to raise later
    from ..exceptions import ExceptionWrapper

    _exc = ExceptionWrapper(ImportError("missing `pip install pycollada`"))
    _collada_loaders.update({"dae": _exc, "zae": _exc})
    _collada_exporters["dae"] = _exc



import json
from tempfile import NamedTemporaryFile

from .. import util
from ..exceptions import ExceptionWrapper


def load_dict(file_obj, **kwargs):
    """
    Load multiple input types into kwargs for a Trimesh constructor.
    Tries to extract keys:
    'faces'
    'vertices'
    'face_normals'
    'vertex_normals'

    Parameters
    ----------
    file_obj : dict
    accepts multiple forms
          -dict: has keys for vertices and faces as (n,3) numpy arrays
          -dict: has keys for vertices/faces (n,3) arrays encoded as dicts/base64
                 with trimesh.util.array_to_encoded/trimesh.util.encoded_to_array
          -str:  json blob as dict with either straight array or base64 values
          -file object: json blob of dict
    file_type: not used

    Returns
    -----------
    loaded: dict with keys
            -vertices: (n,3) float
            -faces:    (n,3) int
            -face_normals: (n,3) float (optional)
    """
    if file_obj is None:
        raise ValueError("file_obj passed to load_dict was None!")
    if util.is_instance_named(file_obj, "Trimesh"):
        return file_obj
    if isinstance(file_obj, str):
        if "{" not in file_obj:
            raise ValueError("Object is not a JSON encoded dictionary!")
        file_obj = json.loads(file_obj.decode("utf-8"))
    elif util.is_file(file_obj):
        file_obj = json.load(file_obj)

    # what shape should the file_obj be to be usable
    mesh_file_obj = {
        "vertices": (-1, 3),
        "faces": (-1, (3, 4)),
        "face_normals": (-1, 3),
        "face_colors": (-1, (3, 4)),
        "vertex_normals": (-1, 3),
        "vertex_colors": (-1, (3, 4)),
    }

    # now go through file_obj structure and if anything is encoded as base64
    # pull it back into numpy arrays
    if not isinstance(file_obj, dict):
        raise ValueError(f"`{type(file_obj)}` object passed to dict loader!")

    loaded = {}
    file_obj = util.decode_keys(file_obj, "utf-8")
    for key, shape in mesh_file_obj.items():
        if key in file_obj:
            loaded[key] = util.encoded_to_array(file_obj[key])
            if not util.is_shape(loaded[key], shape):
                raise ValueError(
                    "Shape of %s is %s, not %s!",
                    key,
                    str(loaded[key].shape),
                    str(shape),
                )
    if len(loaded) == 0:
        raise ValueError("Unable to extract a mesh from the dict!")

    return loaded


def load_meshio(file_obj, file_type: str, **kwargs):
    """
    Load a meshio-supported file into the kwargs for a Trimesh
    constructor.


    Parameters
    ----------
    file_obj : file object
      Contains a meshio file
    file_type : str
      File extension, aka 'vtk'

    Returns
    ----------
    loaded : dict
      kwargs for Trimesh constructor
    """
    # trimesh "file types" are really filename extensions
    # meshio may return multiple answers for each file extension
    file_formats = meshio.extension_to_filetypes["." + file_type]

    mesh = None
    exceptions = []

    # meshio appears to only support loading by file name so use a tempfile
    with NamedTemporaryFile(suffix=f".{file_type}") as temp:
        temp.write(file_obj.read())
        temp.flush()
        # try the loaders in order
        for file_format in file_formats:
            try:
                mesh = meshio.read(temp.name, file_format=file_format)
                break
            except BaseException as E:
                exceptions.append(str(E))

    if mesh is None:
        raise ValueError("Failed to load file:" + "\n".join(exceptions))

    # save file_obj as kwargs for a trimesh.Trimesh
    result = {}
    # pass kwargs to mesh constructor
    result.update(kwargs)
    # add vertices
    result["vertices"] = mesh.points
    try:
        # add faces
        result["faces"] = mesh.get_cells_type("triangle")
    except BaseException:
        util.log.warning("unable to get faces", exc_info=True)
        result["faces"] = []

    return result


_misc_loaders = {"dict": load_dict, "dict64": load_dict}
_misc_loaders = {}


try:
    import meshio

    # add meshio loaders here
    _meshio_loaders = {k[1:]: load_meshio for k in meshio.extension_to_filetypes.keys()}
    _misc_loaders.update(_meshio_loaders)
except BaseException:
    _meshio_loaders = {}

try:
    import openctm

    _misc_loaders["ctm"] = openctm.load_ctm
except BaseException as E:
    _misc_loaders["ctm"] = ExceptionWrapper(E)



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import gmsh
except ImportError:
    gmsh = None


class GMSHTest(g.unittest.TestCase):
    def test_generate(self):
        # don't test gmsh if not installed
        if gmsh is None:
            return
        # create volume mesh of simple icosahedron
        m = g.trimesh.creation.icosahedron()
        result = g.trimesh.interfaces.gmsh.to_volume(m)
        assert len(result) > 0

    def test_load(self):
        if gmsh is None:
            return
        r = g.trimesh.interfaces.gmsh.load_gmsh(
            g.os.path.join(g.dir_models, "featuretype.STEP")
        )
        assert isinstance(r, dict)
        assert len(g.trimesh.Trimesh(**r).faces) > 0

    def test_to_volume(self):
        if gmsh is None:
            return
        m = g.trimesh.creation.box()

        r = g.trimesh.interfaces.gmsh.to_volume(mesh=m)

        assert isinstance(r, bytes)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class SphericalTests(g.unittest.TestCase):
    def test_spherical(self):
        """
        Convert vectors to spherical coordinates
        """
        # random unit vectors
        v = g.trimesh.unitize(g.random((1000, 3)) - 0.5)
        # (n, 2) angles in radians
        spherical = g.trimesh.util.vector_to_spherical(v)
        # back to unit vectors
        v2 = g.trimesh.util.spherical_to_vector(spherical)

        assert g.np.allclose(v, v2)


class HemisphereTests(g.unittest.TestCase):
    def test_hemisphere(self):
        for dimension in [2, 3]:
            # random unit vectors
            v = g.trimesh.unitize(g.random((10000, dimension)) - 0.5)

            # add some on- axis points
            v[:dimension] = g.np.eye(dimension)

            # stack vector and negative vector into one soup
            v = g.np.column_stack((v, -v)).reshape((-1, dimension))

            resigned = g.trimesh.util.vector_hemisphere(v)

            # after resigning, negative vectors should equal positive
            check = (
                abs(g.np.diff(resigned.reshape((-1, 2, dimension)), axis=1).sum(axis=2))
                < g.trimesh.constants.tol.zero
            ).all()
            assert check

            a, s = g.trimesh.util.vector_hemisphere(v, return_sign=True)
            assert g.np.allclose(v, a * s.reshape((-1, 1)))


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import fcl
except BaseException:
    fcl = None


class CollisionTest(g.unittest.TestCase):
    def test_collision(self):
        # Ensure that FCL is importable
        if fcl is None:
            g.log.warning("skipping FCL tests: not installed")
            return

        cube = g.get_mesh("unit_cube.STL")

        tf1 = g.np.eye(4)
        tf1[:3, 3] = g.np.array([5, 0, 0])

        tf2 = g.np.eye(4)
        tf2[:3, 3] = g.np.array([-5, 0, 0])

        # Test one-to-many collision checking
        m = g.trimesh.collision.CollisionManager()
        m.add_object("cube0", cube)
        m.add_object("cube1", cube, tf1)

        ret = m.in_collision_single(cube)
        assert ret is True

        ret, names, data = m.in_collision_single(
            cube, tf1, return_names=True, return_data=True
        )

        assert ret is True
        for c in data:
            assert g.np.allclose(c.point, g.np.array([5.0, -0.5, 0.5]))
            assert g.np.isclose(c.depth, 1.0)
            assert g.np.allclose(c.normal, g.np.array([-1.0, 0.0, 0.0]))

        if "cube1" not in names:
            g.log.debug("\n\n", m._objs.keys(), names)
        assert "cube1" in names

        ret, names, data = m.in_collision_single(
            cube, tf2, return_names=True, return_data=True
        )
        assert ret is False
        assert len(names) == 0
        assert all(len(i.point) == 3 for i in data)

        # Test internal collision checking and object
        # addition/removal/modification
        ret = m.in_collision_internal()
        assert ret is False

        m.add_object("cube2", cube, tf1)
        ret, names = m.in_collision_internal(return_names=True)
        assert ret is True
        assert ("cube1", "cube2") in names
        assert ("cube0", "cube1") not in names
        assert ("cube2", "cube1") not in names

        m.set_transform("cube2", tf2)
        ret = m.in_collision_internal()
        assert ret is False

        m.set_transform("cube2", tf1)
        ret = m.in_collision_internal()
        assert ret is True

        m.remove_object("cube2")
        ret = m.in_collision_internal()
        assert ret is False

        # Test manager-to-manager collision checking
        m = g.trimesh.collision.CollisionManager()
        m.add_object("cube0", cube)
        m.add_object("cube1", cube, tf1)

        n = g.trimesh.collision.CollisionManager()
        n.add_object("cube0", cube, tf2)

        ret = m.in_collision_other(n)
        assert ret is False

        n.add_object("cube3", cube, tf1)

        ret = m.in_collision_other(n)
        assert ret is True

        ret, names = m.in_collision_other(n, return_names=True)
        assert ret is True
        assert ("cube1", "cube3") in names
        assert ("cube3", "cube1") not in names

    def test_random_spheres(self):
        if fcl is None:
            g.log.warning("skipping FCL tests: not installed")
            return

        # check to see if a scene with a bunch of random
        # spheres
        spheres = [
            g.trimesh.creation.icosphere(radius=i[0]).apply_translation(i[1:] * 100)
            for i in g.random((1000, 4))
        ]
        scene = g.trimesh.Scene(spheres)
        manager, _ = g.trimesh.collision.scene_to_collision(scene)
        collides = manager.in_collision_internal()
        assert isinstance(collides, bool)

    def test_distance(self):
        if fcl is None:
            g.log.warning("skipping FCL tests: not installed")
            return

        cube = g.get_mesh("unit_cube.STL")

        tf1 = g.np.eye(4)
        tf1[:3, 3] = g.np.array([5, 0, 0])

        tf2 = g.np.eye(4)
        tf2[:3, 3] = g.np.array([-5, 0, 0])

        tf3 = g.np.eye(4)
        tf3[:3, 3] = g.np.array([2, 0, 0])

        tf4 = g.np.eye(4)
        tf4[:3, 3] = g.np.array([-2, 0, 0])

        tf5 = g.np.eye(4)
        tf5[:3, 3] = g.np.array([5.75, 0, 0])

        # Test one-to-many distance checking
        m = g.trimesh.collision.CollisionManager()
        m.add_object("cube1", cube, tf1)

        dist = m.min_distance_single(cube)
        assert g.np.isclose(dist, 4.0)

        dist = m.min_distance_single(cube, tf5)
        assert g.np.isclose(dist, -0.25)

        dist, name = m.min_distance_single(cube, return_name=True)
        assert g.np.isclose(dist, 4.0)
        assert name == "cube1"

        m.add_object("cube2", cube, tf2)

        dist, name = m.min_distance_single(cube, tf3, return_name=True)
        assert g.np.isclose(dist, 2.0)
        assert name == "cube1"

        dist, name = m.min_distance_single(cube, tf4, return_name=True)
        assert g.np.isclose(dist, 2.0)
        assert name == "cube2"

        # Test internal distance checking and object
        # addition/removal/modification
        dist = m.min_distance_internal()
        assert g.np.isclose(dist, 9.0)

        dist, names = m.min_distance_internal(return_names=True)
        assert g.np.isclose(dist, 9.0)
        assert names == ("cube1", "cube2")

        m.add_object("cube3", cube, tf3)

        dist, names = m.min_distance_internal(return_names=True)
        assert g.np.isclose(dist, 2.0)
        assert names == ("cube1", "cube3")

        m.set_transform("cube3", tf4)

        dist, names = m.min_distance_internal(return_names=True)
        assert g.np.isclose(dist, 2.0)
        assert names == ("cube2", "cube3")

        # Test manager-to-manager distance checking
        m = g.trimesh.collision.CollisionManager()
        m.add_object("cube0", cube)
        m.add_object("cube1", cube, tf1)

        n = g.trimesh.collision.CollisionManager()
        n.add_object("cube4", cube, tf2)

        dist, names, data = m.min_distance_other(n, return_names=True, return_data=True)
        assert g.np.isclose(dist, 4.0)
        assert names == ("cube0", "cube4")
        assert g.np.isclose(
            g.np.linalg.norm(data.point(names[0]) - data.point(names[1])), dist
        )

        n.add_object("cube5", cube, tf4)

        dist, names, data = m.min_distance_other(n, return_names=True, return_data=True)
        assert g.np.isclose(dist, 1.0)
        assert names == ("cube0", "cube5")
        assert g.np.isclose(
            g.np.linalg.norm(data.point(names[0]) - data.point(names[1])), dist
        )

    def test_scene(self):
        if fcl is None:
            return
        scene = g.get_mesh("cycloidal.3DXML")
        manager, objects = g.trimesh.collision.scene_to_collision(scene)

        assert manager.in_collision_internal()
        assert objects is not None


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_rabbit():
    # A BSD-licensed test model from pyglet
    # it has mixed triangles, quads, and 16 element faces -_-
    # this should test the non-vectorized load path
    m = g.get_mesh("rabbit.obj")
    assert len(m.faces) == 1252
    rec = g.roundtrip(m.export(file_type="obj"), file_type="obj")
    assert g.np.isclose(m.area, rec.area)


def test_no_img():
    # sometimes people use the `vt` parameter for arbitrary
    # vertex attributes and thus want UV coordinates even
    # if there is no texture image
    m = g.get_mesh("noimg.obj")
    assert m.visual.uv.shape == (len(m.vertices), 2)
    # make sure UV coordinates are in range 0.0 - 1.0
    assert m.visual.uv.max() < (1 + 1e-5)
    assert m.visual.uv.min() > -1e-5
    # check to make sure it's not all zeros
    assert g.np.ptp(m.visual.uv) > 0.5
    rec = g.roundtrip(m.export(file_type="obj"), file_type="obj")
    assert g.np.isclose(m.area, rec.area)


def test_keep_unreferenced():
    # try loading a short mesh with 2 vertices, one of which is referenced
    m = g.trimesh.load(
        g.trimesh.util.wrap_as_stream("o mesh\nv 1 1 1\nv 1 1 2\nf 1 1 1"),
        file_type="obj",
        process=False,
        maintain_order=True,
    )

    assert g.np.allclose(m.faces[0], [0, 0, 0])
    assert g.np.allclose(m.vertices, [[1, 1, 1], [1, 1, 2]])


def test_trailing():
    # test files with texture and trailing slashes
    m = g.get_mesh("jacked.obj")
    assert len(m.visual.uv) == len(m.vertices)
    rec = g.roundtrip(m.export(file_type="obj"), file_type="obj")
    assert g.np.isclose(m.area, rec.area)


def test_obj_groups():
    # a wavefront file with groups defined
    mesh = g.get_mesh("groups.obj")

    # make sure some data got loaded
    assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
    assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

    # make sure groups are the right length
    # TODO: we do not support face groups now
    # assert len(mesh.metadata['face_groups']) == len(mesh.faces)

    # check to make sure there is signal not just zeros
    # assert g.np.ptp(mesh.metadata['face_groups']) > 0


def test_obj_negative_indices():
    # a wavefront file with negative indices
    mesh = g.get_mesh("negative_indices.obj", merge_tex=True, merge_norm=True)

    # make sure some data got loaded
    assert mesh.faces.shape == (12, 3)
    assert mesh.vertices.shape == (8, 3)


def test_obj_quad():
    mesh = g.get_mesh("quadknot.obj", merge_tex=True)
    # make sure some data got loaded
    assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
    assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

    assert mesh.is_watertight
    assert mesh.is_winding_consistent
    rec = g.roundtrip(mesh.export(file_type="obj"), file_type="obj")
    assert g.np.isclose(mesh.area, rec.area)


def test_obj_multiobj():
    # test a wavefront file with multiple objects in the same file
    scene = g.get_mesh("two_objects.obj", split_object=True, group_material=False)
    assert len(scene.geometry) == 2

    for mesh in scene.geometry.values():
        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

        assert mesh.is_watertight
        assert mesh.is_winding_consistent


def test_obj_split_attributes():
    # test a wavefront file where pos/uv/norm have different indices
    # and where multiple objects share vertices
    # Note 'process=False' to avoid merging vertices
    scene = g.get_mesh(
        "joined_tetrahedra.obj",
        process=False,
        split_object=True,
        group_material=False,
    )

    assert len(scene.geometry) == 2

    geom = list(scene.geometry.values())

    assert g.trimesh.util.is_shape(geom[0].faces, (4, 3))
    assert g.trimesh.util.is_shape(geom[0].vertices, (9, 3))
    assert g.trimesh.util.is_shape(geom[1].faces, (4, 3))
    assert g.trimesh.util.is_shape(geom[1].vertices, (9, 3))


def test_obj_simple_order():
    # test a simple wavefront model without split indexes
    # and make sure we don't reorder vertices unnecessarily
    file_name = g.os.path.join(g.dir_models, "cube.OBJ")
    # load a simple OBJ file without merging vertices
    m = g.trimesh.load(file_name, process=False)
    # use trivial loading to compare with fancy performant one
    with open(file_name) as f:
        f, v, _vt = simple_load(f.read())
    # trimesh loader should return the same face order
    assert g.np.allclose(f, m.faces)
    assert g.np.allclose(v, m.vertices)


def test_order_tex():
    # test a simple wavefront model without split indexes
    # and make sure we don't reorder vertices unnecessarily
    file_name = g.os.path.join(g.dir_models, "fuze.obj")
    # load a simple OBJ file without merging vertices
    m = g.trimesh.load(file_name, process=False, maintain_order=True)
    # use trivial loading to compare with fancy performant one
    with open(file_name) as f:
        f, v, _vt = simple_load(f.read())
    # trimesh loader should return the same face order
    assert g.np.allclose(f, m.faces)
    assert g.np.allclose(v, m.vertices)


def test_obj_compressed():
    mesh = g.get_mesh("cube_compressed.obj", process=False)
    assert mesh._cache.cache["vertex_normals"].shape == mesh.vertices.shape
    assert g.np.allclose(g.np.abs(mesh.vertex_normals).sum(axis=1), 1.0)


def test_vertex_color():
    # get a box mesh
    mesh = g.trimesh.creation.box()
    # set each vertex to a unique random color
    mesh.visual.vertex_colors = [
        g.trimesh.visual.random_color() for _ in range(len(mesh.vertices))
    ]
    # export and then reload the file as OBJ
    rec = g.trimesh.load(
        g.trimesh.util.wrap_as_stream(mesh.export(file_type="obj")), file_type="obj"
    )
    # assert colors have survived the export cycle
    assert (mesh.visual.vertex_colors == rec.visual.vertex_colors).all()


def test_single_vn():
    """
    Make sure files with a single VN load.
    """
    m = g.get_mesh("singlevn.obj")
    assert len(m.vertices) > 0
    assert len(m.faces) > 0


def test_polygon_faces():
    m = g.get_mesh("polygonfaces.obj")
    assert len(m.vertices) > 0
    assert len(m.faces) > 0


def test_faces_not_enough_indices():
    m = g.get_mesh("notenoughindices.obj")
    assert len(m.vertices) > 0
    assert len(m.faces) == 1


def test_export_path():
    m = g.get_mesh("fuze.obj")
    g.check_fuze(m)
    assert m._cache.cache["vertex_normals"].shape == m.vertices.shape
    with g.TemporaryDirectory() as d:
        file_path = g.os.path.join(d, "fz.obj")
        m.export(file_path)
        r = g.trimesh.load(file_path)
        g.check_fuze(r)


def test_skip_mtl():
    # not loading materials should produce a trivial texture
    m_tex = g.get_mesh("fuze.obj")
    m_tex_size = m_tex.visual.material.image.size

    m_notex = g.get_mesh("fuze.obj", skip_materials=True)
    m_notex_size = m_notex.visual.material.image.size

    assert m_tex_size != m_notex_size


def test_mtl():
    # get a mesh with texture
    m = g.get_mesh("fuze.obj")
    # export the mesh including data
    obj, data = g.trimesh.exchange.export.export_obj(m, return_texture=True)
    with g.TemporaryDirectory() as path:
        # where is the OBJ file going to be saved
        obj_path = g.os.path.join(path, "test.obj")
        with open(obj_path, "w") as f:
            f.write(obj)
        # save the MTL and images
        for k, v in data.items():
            with open(g.os.path.join(path, k), "wb") as f:
                f.write(v)
        # reload the mesh from the export
        rec = g.trimesh.load(obj_path)
    # make sure loaded image is the same size as the original
    assert rec.visual.material.image.size == m.visual.material.image.size
    # make sure the faces are the same size
    assert rec.faces.shape == m.faces.shape


def test_scene():
    s = g.get_mesh("cycloidal.3DXML")

    e = g.trimesh.load(
        g.io_wrap(s.export(file_type="obj")),
        file_type="obj",
        split_object=True,
        group_materials=False,
    )

    assert g.np.isclose(e.area, s.area, rtol=0.01)


def test_edge_cases():
    # a mesh with some NaN colors
    n = g.get_mesh("nancolor.obj")
    assert n.faces.shape == (12, 3)

    v = g.get_mesh("cubevt.obj")
    assert v.faces.shape == (12, 3)


def test_empty_or_pointcloud():
    # demo files to check
    empty_files = ["obj_empty.obj", "obj_points.obj", "obj_wireframe.obj"]

    for empty_file in empty_files:
        e = g.get_mesh("emptyIO/" + empty_file)

        # create export
        if "empty" in empty_file:
            try:
                export = e.export(file_type="ply")
            except BaseException:
                continue
            raise ValueError("cannot export empty")
        elif "points" in empty_file:
            export = e.export(file_type="ply")
            reconstructed = g.roundtrip(export, file_type="ply")

            # result should be a point cloud instance
            assert isinstance(e, g.trimesh.PointCloud)
            assert hasattr(e, "vertices")
            # point cloud export should contain vertices
            assert isinstance(reconstructed, g.trimesh.PointCloud)
            assert hasattr(reconstructed, "vertices")


def test_backslash_continuation_character():
    # an obj file with \ (backslash) line continuation characters
    m = g.get_mesh("wallhole.obj")
    assert m.faces.shape == (66, 3)


def test_no_uv():
    mesh = g.get_mesh("box.obj")
    rec = g.roundtrip(mesh.export(file_type="obj"), file_type="obj")
    assert g.np.isclose(mesh.area, rec.area)


def test_no_uv_but_mtl():
    sphere = g.trimesh.creation.uv_sphere()
    sphere.visual = g.trimesh.visual.TextureVisuals(
        uv=None, material=g.trimesh.visual.material.empty_material()
    )
    output = sphere.export("sphere.obj")
    assert "usemtl" in output


def test_chair():
    mesh = next(iter(g.get_mesh("chair.zip").geometry.values()))

    # this model comes with vertex normals
    assert "vertex_normals" in mesh._cache
    assert g.np.allclose(1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
    mesh.apply_scale(0.46377314288075433)
    assert "vertex_normals" in mesh._cache
    assert g.np.allclose(1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
    assert "vertex_normals" in mesh._cache
    mesh._cache.clear()
    assert "vertex_normals" not in mesh._cache
    # if we recomputed now, the degenerate faces
    # would lead some of these vertex normals to be zero
    # assert g.np.allclose(
    #    1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))


def test_multi_nodupe():
    s = g.get_mesh("forearm.zae")
    obj, mtl = g.trimesh.exchange.obj.export_obj(
        s, include_color=True, include_texture=True, return_texture=True
    )
    # should be using one material file
    assert obj.count("mtllib") == 1
    assert "mtllib material.mtl" in obj
    # should be specifying 5 materials
    assert obj.count("usemtl") == 5

    # this file has only the properties (no images)
    assert len(mtl) == 1
    mtl_names = [
        L.strip().split()[-1].strip()
        for L in mtl["material.mtl"].decode("utf-8").split("\n")
        if "newmtl" in L
    ]
    # there should be 5 unique material names
    assert len(set(mtl_names)) == 5


def test_mtl_color_roundtrip():
    # create a mesh with a simple material
    m = g.trimesh.creation.box()
    m.visual = m.visual.to_texture()
    # set each color component to a unique value
    colors = [g.trimesh.visual.color.random_color() for _ in range(3)]
    m.visual.material.ambient = colors[0]
    m.visual.material.specular = colors[1]
    m.visual.material.diffuse = colors[2]
    m.visual.material.glossiness = 0.52622

    with g.TemporaryDirectory() as d:
        # exporting by filename will automatically
        # create a FilePathResolver which writes the
        # `mtl` file to the same directory
        file_name = g.os.path.join(d, "hi.obj")
        m.export(file_name)
        # reload the export by file name
        r = g.trimesh.load(file_name)

    # these values should have survived the roundtrip
    assert g.np.allclose(m.visual.material.ambient, r.visual.material.ambient)
    assert g.np.allclose(m.visual.material.specular, r.visual.material.specular)
    assert g.np.allclose(m.visual.material.diffuse, r.visual.material.diffuse)
    assert g.np.isclose(m.visual.material.glossiness, r.visual.material.glossiness)


def test_scene_export_material_name():
    s = g.get_mesh("fuze.obj", force="scene")

    g.log.warning(s.geometry)

    dummy = "fuxx"
    s.geometry["fuze.obj"].visual.material.name = dummy

    r = g.trimesh.resolvers.ZipResolver()
    r["model.obj"] = s.export(file_type="obj", mtl_name="mystuff.mtl", resolver=r)

    mtl = r["mystuff.mtl"].decode("utf-8")
    assert mtl.count("newmtl") == 1
    assert f"newmtl {dummy}" in mtl
    assert f"{dummy}.jpeg" in r


def test_compound_scene_export():
    # generate a mesh with multiple textures
    a = g.get_mesh("BoxTextured.glb")
    a = a.scaled(1.0 / a.extents.max())
    a.apply_translation(-a.bounds[0])

    b = g.get_mesh("fuze.obj").scene()
    b = b.scaled(1.0 / b.extents.max())
    b.apply_translation(-b.bounds[0] + [2, 0, 0])

    d = next(iter(b.copy().geometry.values()))
    d.apply_translation([-1, 0, 0])
    assert hash(d.visual.material) == hash(b.geometry["fuze.obj"].visual.material)

    # should change the material hash
    d.visual.material.glossiness = 0.1
    assert hash(d.visual.material) != hash(b.geometry["fuze.obj"].visual.material)

    # generate a compound scene
    c = a + b + d
    for i in c.geometry.values():
        # name all the materials the same thing
        i.visual.material.name = "material_0"

    # export the compound scene
    obj, mtl = c.export(file_type="obj", return_texture=True)
    # there should be exactly one mtllib referenced
    assert obj.count("mtllib") == 1
    assert obj.count("usemtl") == 3

    # should be one texture image for each of 3
    # plus the `.mtl` file itself
    # if we had image-hash-deduplication this should
    # be changed to 3 as the image for `b` and `d` are the same
    assert len(mtl) == 4

    # get the material names specified

    mtl_names = [
        L.strip().split()[-1].strip()
        for L in mtl["material.mtl"].decode("utf-8").split("\n")
        if "newmtl" in L
    ]
    # there should be 3 unique material names
    assert len(set(mtl_names)) == 3

    # now reload the compound scene
    t = g.trimesh.load(
        file_obj=g.trimesh.util.wrap_as_stream(obj),
        file_type="obj",
        resolver=g.trimesh.resolvers.ZipResolver(mtl),
        group_material=False,
        split_object=True,
    )
    # these names should match eventually
    assert len(t.geometry.keys()) == len(c.geometry.keys())
    assert g.np.isclose(t.area, c.area)


def test_face_parsing_in_group_names():
    # Checks that an obj with a g tag containinig a face like name (an 'f '
    # followed by three space separated text chunks, ex: f 1 2 3) does load
    # properly
    m = g.get_mesh("face_in_group_name.obj")
    assert len(m.vertices) == 1


def test_face_parsing_in_group_names_with_object_tag():
    # Checks that an obj with a g tag in the middle of a file,
    # containinig a face like name (an 'f ' followed by three
    # space separated text chunks, ex: f 1 2 3), does load properly
    m = g.get_mesh("face_in_group_name_mid_file.obj")
    assert len(m.vertices) == 5
    assert len(m.faces) == 2


def test_chunk_parsing_with_no_faces_but_with_f_in_chunk():
    # Checks that a chunk with no faces but with 'f ' in it loads properly
    m = g.get_mesh("obj_with_no_face_in_chunk.obj")
    assert len(m.vertices) == 3
    assert len(m.faces) == 1


def test_export_normals():
    m = g.trimesh.creation.box()
    assert "vertex_normals" not in m._cache.cache

    e = m.export(file_type="obj", include_normals=None)
    assert "vn " not in e
    e = m.export(file_type="obj", include_normals=True)
    # should have included normals
    assert "vn " in e
    # should have forced generation of normals
    assert "vertex_normals" in m._cache.cache

    # now that they're in cache include_normals=None should get them
    e = m.export(file_type="obj", include_normals=None)
    assert "vn " in e

    # or skip them if explicitly asked
    e = m.export(file_type="obj", include_normals=False)
    assert "vn " not in e


def test_export_mtl_args():
    mesh = g.trimesh.creation.box()
    # check for a crash with no materials defined
    _a, _b = g.trimesh.exchange.obj.export_obj(
        mesh, return_texture=True, mtl_name="hi.mtl"
    )


def simple_load(text):
    # we're going to load faces in a basic text way
    # and compare the order from this method to the
    # trimesh loader, to see if we get the same thing
    # note that trimesh's extremely convoluted string
    # wangling is wildly faster than this
    f = []
    v = []
    vt = []
    for line in str.splitlines(text):
        line = line.strip()
        if len(line) < 2:
            continue
        elif line.startswith("f "):
            if "/" in line:
                f.append([int(i.split("/", 1)[0]) for i in line[1:].strip().split()])
            else:
                f.append(line[1:].strip().split())
        elif line.startswith("v "):
            v.append(line[1:].strip().split())
        elif line.startswith("vt "):
            vt.append(line[2:].strip().split())

    # get faces as basic numpy array
    f = g.np.array(f, dtype=g.np.int64) - 1
    v = g.np.array(v, dtype=g.np.float64)
    vt = g.np.array(vt, dtype=g.np.float64)

    return f, v, vt


def test_material_name():
    m = g.get_mesh("fuze.obj")

    # check that the material used the name in the file
    assert m.visual.material.name == "a-super-duper-material"

    # roundtrip the OBJ through a file path
    # this is easier than making a resolver ourself for the MTL file
    # as it will automatically create the `FilePathResolver` on paths
    with g.TemporaryDirectory() as temp:
        path = g.os.path.join(temp, "mesh.obj")
        m.export(file_obj=path)
        roundtrip = g.trimesh.load_mesh(file_obj=path)

    # material name should have survivied
    assert roundtrip.visual.material.name == "a-super-duper-material"

    g.check_fuze(roundtrip)


def test_obj_quad_uv():
    # check that the quads with UV's get translated
    m = g.get_mesh("quadknot.obj")

    assert m.visual.uv.shape[0] == len(m.vertices)
    assert g.np.isclose(m.visual.uv.max(), 0.1)


if __name__ == "__main__":
    # test_material_name()
    test_obj_quad_uv()



import numpy as np

from ..constants import log
from ..exceptions import ExceptionWrapper
from ..typed import ArrayLike, Number, Optional

try:
    from PIL.Image import Image, fromarray
except BaseException as E:
    Image = ExceptionWrapper(E)
    fromarray = ExceptionWrapper(E)


def specular_to_pbr(
    specularFactor: Optional[ArrayLike] = None,
    glossinessFactor: Optional[Number] = None,
    specularGlossinessTexture: Optional["Image"] = None,
    diffuseTexture: Optional["Image"] = None,
    diffuseFactor: Optional[ArrayLike] = None,
    **kwargs,
) -> dict:
    """
    Convert the KHR_materials_pbrSpecularGlossiness to a
    metallicRoughness visual.

    Parameters
    -----------
    specularFactor : list[float]
        Specular color values. Ignored if specularGlossinessTexture
        is present and defaults to [1.0, 1.0, 1.0].
    glossinessFactor : float
        glossiness factor in range [0, 1], scaled
        specularGlossinessTexture if present.
        Defaults to 1.0.
    specularGlossinessTexture : PIL.Image
        Texture with 4 color channels. With [0,1,2] representing
        specular RGB and 3 glossiness.
    diffuseTexture : PIL.Image
        Texture with 4 color channels. With [0,1,2] representing diffuse
        RGB and 3 opacity.
    diffuseFactor: float
        Diffuse RGBA color. scales diffuseTexture if present.
        Defaults to [1.0, 1.0, 1.0, 1.0].

    Returns
    ----------
    kwargs : dict
      Constructor args for a PBRMaterial object.
      Containing:
        - either baseColorTexture or baseColorFactor
        - either metallicRoughnessTexture or metallicFactor and roughnessFactor
    """
    # based on:
    # https://github.com/KhronosGroup/glTF/blob/89427b26fcac884385a2e6d5803d917ab5d1b04f/extensions/2.0/Archived/KHR_materials_pbrSpecularGlossiness/examples/convert-between-workflows-bjs/js/babylon.pbrUtilities.js#L33-L64

    if isinstance(Image, ExceptionWrapper):
        log.debug("unable to convert specular-glossy material without pillow!")
        result = {}
        if isinstance(diffuseTexture, dict):
            result["baseColorTexture"] = diffuseTexture
        if diffuseFactor is not None:
            result["baseColorFactor"] = diffuseFactor
        return result

    dielectric_specular = np.array([0.04, 0.04, 0.04], dtype=np.float32)
    epsilon = 1e-6

    def solve_metallic(diffuse, specular, one_minus_specular_strength):
        if isinstance(specular, float) and specular < dielectric_specular[0]:
            return 0.0

        if len(diffuse.shape) == 2:
            diffuse = diffuse[..., None]
        if len(specular.shape) == 2:
            specular = specular[..., None]

        a = dielectric_specular[0]
        b = (
            diffuse * one_minus_specular_strength / (1.0 - dielectric_specular[0])
            + specular
            - 2.0 * dielectric_specular[0]
        )
        c = dielectric_specular[0] - specular
        D = b * b - 4.0 * a * c
        D = np.clip(D, epsilon, None)
        metallic = np.clip((-b + np.sqrt(D)) / (2.0 * a), 0.0, 1.0)
        if isinstance(metallic, np.ndarray):
            metallic[specular < dielectric_specular[0]] = 0.0
        return metallic

    def get_perceived_brightness(rgb):
        return np.sqrt(np.dot(rgb[..., :3] ** 2, [0.299, 0.587, 0.114]))

    def toPIL(img, mode=None):
        if isinstance(img, Image):
            return img
        if img.dtype == np.float32 or img.dtype == np.float64:
            img = (np.clip(img, 0.0, 1.0) * 255.0).astype(np.uint8)
        return fromarray(img)

    def get_float(val):
        if isinstance(val, float):
            return val
        if isinstance(val, np.ndarray) and len(val.shape) == 1:
            return val[0]
        return val.tolist()

    def get_diffuse(diffuseFactor, diffuseTexture):
        diffuseFactor = (
            diffuseFactor if diffuseFactor is not None else [1.0, 1.0, 1.0, 1.0]
        )
        diffuseFactor = np.array(diffuseFactor, dtype=np.float32)

        if diffuseTexture is not None:
            if diffuseTexture.mode == "BGR":
                diffuseTexture = diffuseTexture.convert("RGB")
            elif diffuseTexture.mode == "BGRA":
                diffuseTexture = diffuseTexture.convert("RGBA")

            diffuse = np.array(diffuseTexture) / 255.0
            # diffuseFactor must be applied to linear scaled colors .
            # Sometimes, diffuse texture is only 2 channels, how do we know
            # if they are encoded sRGB or linear?
            diffuse = convert_texture_srgb2lin(diffuse)

            if len(diffuse.shape) == 2:
                diffuse = diffuse[..., None]
            if diffuse.shape[-1] == 1:
                diffuse = diffuse * diffuseFactor
            elif diffuse.shape[-1] == 2:
                alpha = diffuse[..., 1:2]
                diffuse = diffuse[..., :1] * diffuseFactor
                if diffuseFactor.shape[-1] == 3:
                    # this should actually not happen, but it seems like many materials are not complying with the spec
                    diffuse = np.concatenate([diffuse, alpha], axis=-1)
                else:
                    diffuse[..., -1:] *= alpha
            elif diffuse.shape[-1] == diffuseFactor.shape[-1]:
                diffuse = diffuse * diffuseFactor
            elif diffuse.shape[-1] == 3 and diffuseFactor.shape[-1] == 4:
                diffuse = (
                    np.concatenate([diffuse, np.ones_like(diffuse[..., :1])], axis=-1)
                    * diffuseFactor
                )
            else:
                log.warning(
                    "`diffuseFactor` and `diffuseTexture` have incompatible shapes: "
                    + f"{diffuseFactor.shape} and {diffuse.shape}"
                )
        else:
            diffuse = diffuseFactor if diffuseFactor is not None else [1, 1, 1, 1]
            diffuse = np.array(diffuse, dtype=np.float32)
        return diffuse

    def get_specular_glossiness(
        specularFactor, glossinessFactor, specularGlossinessTexture
    ):
        if specularFactor is None:
            specularFactor = [1.0, 1.0, 1.0]
        specularFactor = np.array(specularFactor, dtype=np.float32)
        if glossinessFactor is None:
            glossinessFactor = 1.0
        glossinessFactor = np.array([glossinessFactor], dtype=np.float32)

        # specularGlossinessTexture should be a texture with 4 channels,
        # 3 sRGB channels for specular and 1 linear channel for glossiness.
        # in practice, it can also have just 1, 2, or 3 channels which are then to
        # be multiplied with the provided factors

        if specularGlossinessTexture is not None:
            if specularGlossinessTexture.mode == "BGR":
                specularGlossinessTexture = specularGlossinessTexture.convert("RGB")
            elif specularGlossinessTexture.mode == "BGRA":
                specularGlossinessTexture = specularGlossinessTexture.convert("RGBA")

            specularGlossinessTexture = np.array(specularGlossinessTexture) / 255.0
            specularTexture, glossinessTexture = None, None

            if len(specularGlossinessTexture.shape) == 2:
                # use the one channel as a multiplier for specular and glossiness
                specularTexture = glossinessTexture = specularGlossinessTexture.reshape(
                    specularGlossinessTexture.shape[0],
                    specularGlossinessTexture.shape[1],
                    1,
                )
            elif specularGlossinessTexture.shape[-1] == 1:
                # use the one channel as a multiplier for specular and glossiness
                specularTexture = glossinessTexture = specularGlossinessTexture[
                    ..., np.newaxis
                ]
            elif specularGlossinessTexture.shape[-1] == 3:
                # all channels are specular, glossiness is only a factor
                specularTexture = specularGlossinessTexture[..., :3]
            elif specularGlossinessTexture.shape[-1] == 2:
                # first channel is specular, last channel is glossiness
                specularTexture = specularGlossinessTexture[..., :1]
                glossinessTexture = specularGlossinessTexture[..., 1:2]
            elif specularGlossinessTexture.shape[-1] == 4:
                # first 3 channels are specular, last channel is glossiness
                specularTexture = specularGlossinessTexture[..., :3]
                glossinessTexture = specularGlossinessTexture[..., 3:]

            if specularTexture is not None:
                # specular texture channels are sRGB
                specularTexture = convert_texture_srgb2lin(specularTexture)
                specular = specularTexture * specularFactor
            else:
                specular = specularFactor

            if glossinessTexture is not None:
                # glossiness texture channel is linear
                glossiness = glossinessTexture * glossinessFactor
            else:
                glossiness = glossinessFactor

            one_minus_specular_strength = 1.0 - np.max(specular, axis=-1, keepdims=True)
        else:
            specular = specularFactor if specularFactor is not None else [1.0, 1.0, 1.0]
            specular = np.array(specular, dtype=np.float32)
            glossiness = glossinessFactor if glossinessFactor is not None else 1.0
            glossiness = np.array(glossiness, dtype=np.float32)
            one_minus_specular_strength = 1.0 - max(specular[:3])

        return specular, glossiness, one_minus_specular_strength

    if diffuseTexture is not None and specularGlossinessTexture is not None:
        # reshape to the size of the largest texture
        max_shape = tuple(
            max(diffuseTexture.size[i], specularGlossinessTexture.size[i])
            for i in range(2)
        )
        if (
            diffuseTexture.size[0] != max_shape[0]
            or diffuseTexture.size[1] != max_shape[1]
        ):
            diffuseTexture = diffuseTexture.resize(max_shape)
        if (
            specularGlossinessTexture.size[0] != max_shape[0]
            or specularGlossinessTexture.size[1] != max_shape[1]
        ):
            specularGlossinessTexture = specularGlossinessTexture.resize(max_shape)

    def srgb2lin(s):
        """
        Converts sRGB color values to linear color values.
        See: https://entropymine.com/imageworsener/srgbformula/
        """

        mask = s <= 0.0404482362771082
        lin = np.empty_like(s)
        lin[mask] = s[mask] / 12.92
        lin[~mask] = np.power(((s[~mask] + 0.055) / 1.055), 2.4)
        return lin

    def convert_texture_srgb2lin(texture):
        """
        Wrapper for srgb2lin that converts color values from sRGB to linear.
        If texture has 2 or 4 channels, the last channel (alpha) is left unchanged.
        """
        result = texture.copy()
        color_channels = result.shape[-1]
        # only scale the color channels, not the alpha channel
        if color_channels == 4 or color_channels == 2:
            color_channels -= 1
        result[..., :color_channels] = srgb2lin(result[..., :color_channels])
        return result

    def lin2srgb(lin):
        """
        Converts linear color values to sRGB color values.
        See: https://entropymine.com/imageworsener/srgbformula/
        """
        s = np.empty_like(lin)
        mask = lin > 0.00313066844250063
        s[mask] = 1.055 * np.power(lin[mask], (1.0 / 2.4)) - 0.055
        s[~mask] = 12.92 * lin[~mask]
        return s

    def convert_texture_lin2srgb(texture):
        """
        Wrapper for lin2srgb that converts color values from linear to sRGB.
        If texture has 2 or 4 channels, the last channel (alpha) is left unchanged.
        """

        result = texture.copy()
        color_channels = result.shape[-1]
        # only scale the color channels, not the alpha channel
        if color_channels == 4 or color_channels == 2:
            color_channels -= 1
        result[..., :color_channels] = lin2srgb(result[..., :color_channels])
        return result

    diffuse = get_diffuse(diffuseFactor, diffuseTexture)
    specular, glossiness, one_minus_specular_strength = get_specular_glossiness(
        specularFactor, glossinessFactor, specularGlossinessTexture
    )

    metallic = solve_metallic(
        get_perceived_brightness(diffuse),
        get_perceived_brightness(specular),
        one_minus_specular_strength,
    )
    if not isinstance(metallic, np.ndarray):
        metallic = np.array(metallic, dtype=np.float32)

    diffuse_rgb = diffuse[..., :3]

    base_color_from_diffuse = diffuse_rgb * (
        one_minus_specular_strength
        / (1.0 - dielectric_specular[0])
        / np.clip((1.0 - metallic), epsilon, None)
    )
    base_color_from_specular = (specular - dielectric_specular * (1.0 - metallic)) * (
        1.0 / np.clip(metallic, epsilon, None)
    )
    mm = metallic * metallic
    base_color = mm * base_color_from_specular + (1.0 - mm) * base_color_from_diffuse
    base_color = np.clip(base_color, 0.0, 1.0)

    # get opacity
    try:
        if diffuse.shape == (4,):
            # opacity is a single scalar value
            opacity = diffuse[-1]
            if base_color.shape == (3,):
                # simple case with one color and diffuse with opacity
                # add on the opacity from the diffuse color
                base_color = np.append(base_color, opacity)
            elif len(base_color.shape) == 3:
                # stack opacity to match the base color array
                dim = base_color.shape
                base_color = np.dstack(
                    (
                        base_color,
                        np.full(np.prod(dim[:2]), opacity).reshape((dim[0], dim[1], 1)),
                    )
                )
        elif diffuse.shape[-1] == 4:
            opacity = diffuse[..., -1]
            base_color = np.concatenate([base_color, opacity[..., None]], axis=-1)
    except BaseException:
        log.error("unable to get opacity", exc_info=True)

    result = {}
    if len(base_color.shape) > 1:
        # convert back to sRGB
        result["baseColorTexture"] = toPIL(
            convert_texture_lin2srgb(base_color),
            mode=("RGB" if base_color.shape[-1] == 3 else "RGBA"),
        )
    else:
        result["baseColorFactor"] = base_color.tolist()

    if len(metallic.shape) > 1 or len(glossiness.shape) > 1:
        if len(glossiness.shape) == 1:
            glossiness = np.tile(glossiness, (metallic.shape[0], metallic.shape[1], 1))
        if len(metallic.shape) == 1:
            metallic = np.tile(metallic, (glossiness.shape[0], glossiness.shape[1], 1))

        # we need to use RGB textures, because 2 channel textures can cause problems
        result["metallicRoughnessTexture"] = toPIL(
            np.concatenate(
                [np.zeros_like(metallic), 1.0 - glossiness, metallic], axis=-1
            ),
            mode="RGB",
        )
        result["metallicFactor"] = 1.0
        result["roughnessFactor"] = 1.0
    else:
        result["metallicFactor"] = get_float(metallic)
        result["roughnessFactor"] = get_float(1.0 - glossiness)

    return result
